# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: labm8/internal/lockfile.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='labm8/internal/lockfile.proto',
  package='',
  syntax='proto2',
  serialized_options=None,
  serialized_pb=_b('\n\x1dlabm8/internal/lockfile.proto\"\xe4\x01\n\x08LockFile\x12(\n\x10owner_process_id\x18\x01 \x01(\x05R\x0eownerProcessId\x12,\n\x12owner_process_argv\x18\x02 \x01(\tR\x10ownerProcessArgv\x12:\n\x1a\x64\x61te_acquired_utc_epoch_ms\x18\x03 \x01(\x03R\x16\x64\x61teAcquiredUtcEpochMs\x12%\n\x0eowner_hostname\x18\x04 \x01(\tR\rownerHostname\x12\x1d\n\nowner_user\x18\x05 \x01(\tR\townerUser')
)




_LOCKFILE = _descriptor.Descriptor(
  name='LockFile',
  full_name='LockFile',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='owner_process_id', full_name='LockFile.owner_process_id', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='ownerProcessId', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='owner_process_argv', full_name='LockFile.owner_process_argv', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='ownerProcessArgv', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='date_acquired_utc_epoch_ms', full_name='LockFile.date_acquired_utc_epoch_ms', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='dateAcquiredUtcEpochMs', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='owner_hostname', full_name='LockFile.owner_hostname', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='ownerHostname', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='owner_user', full_name='LockFile.owner_user', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='ownerUser', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=34,
  serialized_end=262,
)

DESCRIPTOR.message_types_by_name['LockFile'] = _LOCKFILE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

LockFile = _reflection.GeneratedProtocolMessageType('LockFile', (_message.Message,), dict(
  DESCRIPTOR = _LOCKFILE,
  __module__ = 'labm8.internal.lockfile_pb2'
  # @@protoc_insertion_point(class_scope:LockFile)
  ))
_sym_db.RegisterMessage(LockFile)


# @@protoc_insertion_point(module_scope)



# Copyright 2014-2019 Chris Cummins <chrisc.101@gmail.com>.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module contains utility functions for parallelism in python.

The goal of the module is to provide easy to use implementations of typical
parallel workloads, such as data parallel map operations.
"""
import multiprocessing
import subprocess
import typing

from labm8 import app
from labm8 import bazelutil
from labm8 import pbutil
from labm8 import sqlutil

FLAGS = app.FLAGS


class MapWorkerError(EnvironmentError):
  """Resulting error from a _MapWorker that fails."""

  def __init__(self, returncode: int):
    """Create a _MapWorker error.

    Args:
      returncode: The process return code.
    """
    self._returncode = returncode

  def __repr__(self) -> str:
    return f'Command exited with code {self.returncode}'

  @property
  def returncode(self) -> int:
    """Get the return code of the process."""
    return self._returncode


class _MapWorker(object):
  """A work unit for a data parallel workload.

  A _MapWorker executes a command as a subprocess, passes it a protocol buffer,
  and decodes a protocol buffer output.

  This is a helper class created by MapNativeProtoProcessingBinary() and
  returned to the user. It is not to be used by user code.
  """

  def __init__(
      self,
      id: int,
      cmd: typing.List[str],
      input_proto: pbutil.ProtocolBuffer,
  ):
    """Create a map worker.

    Args:
      id: The numeric ID of the map worker.
      cmd: The command to execute, as a list of arguments to subprocess.Popen().
      input_proto: The protocol buffer to pass to the command.
    """
    self._id = id
    self._cmd = cmd
    # We store the input proto in wire format (as a serialized string) rather
    # than as a class object as pickle can get confused by the types.
    # See: https://stackoverflow.com/a/1413299
    self._input_proto: typing.Optional[pbutil.ProtocolBuffer] = None
    self._input_proto_string = input_proto.SerializeToString()
    self._output_proto_string: typing.Optional[str] = None
    self._output_proto: typing.Optional[pbutil.ProtocolBuffer] = None
    self._output_proto_decoded = False
    self._returncode: typing.Optional[int] = None
    self._done = False

  def Run(self) -> None:
    """Execute the process and store the output.

    If the process fails, no exception is raised. The error can be accessed
    using the error() method. After calling this method, SetProtos() *must* be
    called.
    """
    assert not self._done

    # Run the C++ worker process, capturing it's output.
    process = subprocess.Popen(
        self._cmd,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
    )
    # Send the input proto to the C++ worker process.
    # TODO: Add timeout.
    stdout, _ = process.communicate(self._input_proto_string)
    self._returncode = process.returncode
    del self._input_proto_string

    if not process.returncode:
      # Store the C++ binary output in wire format.
      self._output_proto_string = stdout

  def SetProtos(
      self,
      input_proto: pbutil.ProtocolBuffer,
      output_proto_class: typing.Type,
  ) -> None:
    """Set the input protocol buffer, and decode the output protocol buffer.

    This is performed by the SetProtos() method (rather than during Run()) so
    that when pickled, this class contains only basic types, not protocol buffer
    instances.

    Args:
      input_proto: The input protocol buffer message.
      output_proto_class: The protocol buffer class of the output message.
    """
    assert not self._done
    self._done = True
    self._input_proto = input_proto
    # Only parse the output if the worker completed successfully.
    if not self._returncode:
      # Be careful that errors during protocol buffer decoding (e.g.
      # unrecognized fields, conflicting field type/tags) are silently ignored
      # here.
      self._output_proto = output_proto_class.FromString(
          self._output_proto_string,)
    # No need to hand onto the string message any more.
    del self._output_proto_string

  @property
  def id(self):
    """Return the numeric ID of the map worker."""
    return self._id

  def input(self) -> pbutil.ProtocolBuffer:
    """Get the input protocol buffer."""
    assert self._done
    return self._input_proto

  def output(self) -> typing.Optional[pbutil.ProtocolBuffer]:
    """Get the protocol buffer decoded from stdout of the executed binary.

    If the process failed (e.g. not _MapWorker.ok()), None is returned.
    """
    assert self._done
    return self._output_proto

  def error(self) -> typing.Optional[MapWorkerError]:
    """Get the error generated by a failed binary execution.

    If the process succeeded (e.g. _MapWorker.ok()), None is returned.
    """
    if self._returncode:
      return MapWorkerError(self._returncode)

  def ok(self) -> bool:
    """Return whether binary execution succeeded."""
    return not self._returncode


def _RunNativeProtoProcessingWorker(map_worker: _MapWorker) -> _MapWorker:
  """Private helper message to execute Run() method of _MapWorker.

  This is passed to Pool.imap_unordered() as the function to execute for every
  work unit. This is needed because only module-level functions can be pickled.
  """
  map_worker.Run()
  return map_worker


def MapNativeProtoProcessingBinary(
    binary_data_path: str,
    input_protos: typing.List[pbutil.ProtocolBuffer],
    output_proto_class: typing.Type,
    binary_args: typing.Optional[typing.List[str]] = None,
    pool: typing.Optional[multiprocessing.Pool] = None,
    num_processes: typing.Optional[int] = None,
) -> typing.Iterator[_MapWorker]:
  """Run a protocol buffer processing binary over a set of inputs.

  Args:
    binary_data_path: The path of the binary to execute, as provied to
      bazelutil.DataPath().
    input_protos: An iterable list of input protos.
    output_proto_class: The proto class of the output.
    binary_args: An optional list of additional arguments to pass to binaries.
    pool: The multiprocessing pool to use.
    num_processes: The number of processes for the multiprocessing pool.

  Returns:
    A generator of _MapWorker instances. The order is random.
  """
  binary_path = bazelutil.DataPath(binary_data_path)
  binary_args = binary_args or []
  cmd = [str(binary_path)] + binary_args

  # Read all inputs to a list. We need the inputs in a list so that we can
  # map an inputs position in the list to a _MapWorker.id.
  input_protos = list(input_protos)

  # Create the multiprocessing pool to use, if not provided.
  pool = pool or multiprocessing.Pool(processes=num_processes)

  map_worker_iterator = (_MapWorker(i, cmd, input_proto)
                         for i, input_proto in enumerate(input_protos))

  for map_worker in pool.imap_unordered(
      _RunNativeProtoProcessingWorker,
      map_worker_iterator,
  ):
    map_worker.SetProtos(input_protos[map_worker.id], output_proto_class)
    yield map_worker


def MapNativeProcessingBinaries(
    binaries: typing.List[str],
    input_protos: typing.List[pbutil.ProtocolBuffer],
    output_proto_classes: typing.List[typing.Type],
    pool: typing.Optional[multiprocessing.Pool] = None,
    num_processes: typing.Optional[int] = None,
) -> typing.Iterator[_MapWorker]:
  """Run a protocol buffer processing binary over a set of inputs.

  Args:
    binary_data_path: The path of the binary to execute, as provied to
      bazelutil.DataPath().
    input_protos: An iterable list of input protos.
    output_proto_class: The proto class of the output.
    binary_args: An optional list of additional arguments to pass to binaries.
    pool: The multiprocessing pool to use.
    num_processes: The number of processes for the multiprocessing pool.

  Returns:
    A generator of _MapWorker instances. The order is random.
  """
  if not len(binaries) == len(input_protos):
    raise ValueError('Number of binaries does not equal protos')

  cmds = [[bazelutil.DataPath(b)] for b in binaries]

  # Read all inputs to a list. We need the inputs in a list so that we can
  # map an inputs position in the list to a _MapWorker.id.
  input_protos = list(input_protos)
  output_proto_classes = list(output_proto_classes)

  # Create the multiprocessing pool to use, if not provided.
  pool = pool or multiprocessing.Pool(processes=num_processes)

  map_worker_iterator = (_MapWorker(
      id,
      cmd,
      input_proto,
  ) for id, (
      cmd,
      input_proto,
  ) in enumerate(zip(cmds, input_protos)))

  for map_worker in pool.imap_unordered(
      _RunNativeProtoProcessingWorker,
      map_worker_iterator,
  ):
    map_worker.SetProtos(
        input_protos[map_worker.id],
        output_proto_classes[map_worker.id],
    )
    yield map_worker


# Type annotations for MapDatabaseRowBatchProcessor().
WorkUnitType = typing.Callable[[typing.List[typing.Any]], typing.Any]
WorkUnitArgGenerator = typing.Callable[[typing.Any], typing.Any]
ResultCallback = typing.Callable[[typing.Any], None]
BatchCallback = typing.Callable[[int], None]


def MapDatabaseRowBatchProcessor(
    work_unit: WorkUnitType,
    query: sqlutil.Query,
    generate_work_unit_args: WorkUnitArgGenerator = lambda rows: rows,
    work_unit_result_callback: ResultCallback = lambda result: None,
    start_of_batch_callback: BatchCallback = lambda i: None,
    end_of_batch_callback: BatchCallback = lambda i: None,
    batch_size: int = 256,
    rows_per_work_unit: int = 5,
    start_at: int = 0,
    pool: typing.Optional[multiprocessing.Pool] = None,
) -> None:
  """Execute a database row-processesing function in parallel.

  Use this function to orchestrate the parallel execution of a function that
  takes batches of rows from the result set of a query.

  This is equivalent to a serial implementation:

    for row in query:
      work_unit_result_callback(work_unit(generate_work_unit_args)))
    end_of_batch_callback()

  Args:
    work_unit: A function which takes an input a list of the values returned
      by generate_work_unit_args callback, and produces a list of zero or more
      instances of output_table_class.
    query: The query which produces inputs to the work units.
    generate_work_unit_args: A callback which transforms a single result of the
      query into an input to a work unit.
    batch_size:
    rows_per_work_unit:
    start_at:
    pool:

  Returns:
    Foo.
  """

  pool = pool or multiprocessing.Pool()

  i = start_at
  row_batches = sqlutil.OffsetLimitBatchedQuery(query, batch_size=batch_size)

  for batch in row_batches:
    rows_batch = batch.rows
    start_of_batch_callback(i)

    work_unit_args = [
        generate_work_unit_args(rows_batch[i:i + rows_per_work_unit])
        for i in range(0, len(rows_batch), rows_per_work_unit)
    ]

    for result in pool.starmap(work_unit, work_unit_args):
      work_unit_result_callback(result)

    i += len(rows_batch)

    end_of_batch_callback(i)



# Copyright 2014-2019 Chris Cummins <chrisc.101@gmail.com>.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility code for converting between absl logging output and log protos."""
import contextlib
import datetime
import pathlib
import re
import sys
import typing

from absl import logging

from labm8 import app
from labm8 import labdate
from labm8.internal import logging_pb2

FLAGS = app.FLAGS

# A regular expression to match the components of an absl logging prefix. See:
# https://github.com/abseil/abseil-py/blob/e69e200f680a20c50e0e2cd9e74e9850ff69b856/absl/logging/__init__.py#L554-L583
ABSL_LOGGING_LINE_RE = re.compile(
    r'(?P<lvl>[IWEF])(?P<timestamp>\d\d\d\d \d\d:\d\d:\d\d.\d\d\d\d\d\d) '
    r'(?P<thread_id>\d+) (?P<filename>[^:]+):(?P<lineno>\d+)] '
    r'(?P<contents>.*)',)

# Convert a single letter absl logging prefix to a LogRecord.LogLevel. Since
# absl logging uses the same prefix for logging.DEBUG and logging.INFO, this
# conversion is lossy, as LogRecord.DEBUG is never returned.
ABSL_LEVEL_TO_LOG_RECORD_LEVEL = {
    'I': logging_pb2.LogRecord.INFO,
    'W': logging_pb2.LogRecord.WARNING,
    'E': logging_pb2.LogRecord.ERROR,
    'F': logging_pb2.LogRecord.FATAL,
}


def DatetimeFromAbslTimestamp(
    timestamp: str,
    year: int = datetime.datetime.utcnow().year,
) -> datetime.datetime:
  """Convert absl logging timestamp to datetime.

  WARNING: Absl logs do not include the year, so if parsing logs from previous
  years, be sure to set the year argument! The default value assumes the logs
  are from the current year.

  Args:
    timestamp: The string timestamp.
    year: The year, as an integer. E.g. 2019.

  Returns:
    A datetime.
  """
  dt = datetime.datetime.strptime(str(year) + timestamp, '%Y%m%d %H:%M:%S.%f')
  return dt


def ConertAbslLogToProtos(
    logs: str,
    year: int = datetime.datetime.utcnow().year,
) -> typing.List[logging_pb2.LogRecord]:
  """Convert the output of logging with absl logging to LogRecord protos.

  WARNING: Absl logs do not include the year, so if parsing logs from previous
  years, be sure to set the year argument! The default value assumes the logs
  are from the current year.

  Args:
    logs: The output from logging with absl.
    year: The year, as an integer. E.g. 2019.

  Returns:
    A list of LogRecord messages.
  """
  records = []
  starting_match = None
  lines_buffer = []

  def ConvertOne() -> logging_pb2.LogRecord:
    """Convert the current starting_match and lines_buffer into a LogRecord."""
    if starting_match:
      records.append(
          logging_pb2.LogRecord(
              level=ABSL_LEVEL_TO_LOG_RECORD_LEVEL[starting_match.group('lvl')],
              date_utc_epoch_ms=labdate.MillisecondsTimestamp(
                  DatetimeFromAbslTimestamp(
                      starting_match.group('timestamp'),
                      year=year,
                  ),),
              thread_id=int(starting_match.group('thread_id')),
              file_name=starting_match.group('filename'),
              line_number=int(starting_match.group('lineno')),
              message='\n'.join(
                  [starting_match.group('contents')] + lines_buffer,).rstrip(),
          ),)

  for line in logs.split('\n'):
    m = ABSL_LOGGING_LINE_RE.match(line)
    if m:
      ConvertOne()
      starting_match = None
      lines_buffer = []
      starting_match = m
    elif line and not starting_match:
      raise ValueError(f"Failed to parse logging output at line: '{line}'")
    else:
      lines_buffer.append(line)
  ConvertOne()
  return records


def StartTeeLogsToFile(
    program_name: str = None,
    log_dir: str = None,
    file_log_level: int = logging.DEBUG,
) -> None:
  """Log messages to file as well as stderr.

  Args:
    program_name: The name of the program.
    log_dir: The directory to log to.
    file_log_level: The minimum verbosity level to log to file to.

  Raises:
    FileNotFoundError: If the requested log_dir does not exist.
  """
  if not pathlib.Path(log_dir).is_dir():
    raise FileNotFoundError(f"Log directory not found: '{log_dir}'")
  old_verbosity = logging.get_verbosity()
  logging.set_verbosity(file_log_level)
  logging.set_stderrthreshold(old_verbosity)
  logging.get_absl_handler().start_logging_to_file(program_name, log_dir)
  # The Absl logging handler function start_logging_to_file() sets logtostderr
  # to False. Re-enable whatever value it was before the call.
  FLAGS.logtostderr = False


def StopTeeLogsToFile():
  """Stop logging messages to file as well as stderr."""
  logging.get_absl_handler().flush()
  logging.get_absl_handler().stream = sys.stderr
  FLAGS.logtostderr = True


@contextlib.contextmanager
def TeeLogsToFile(
    program_name: str = None,
    log_dir: str = None,
    file_log_level: int = logging.DEBUG,
):
  """Temporarily enable logging to file.

  Args:
    program_name: The name of the program.
    log_dir: The directory to log to.
    file_log_level: The minimum verbosity level to log to file to.
  """
  try:
    StartTeeLogsToFile(program_name, log_dir, file_log_level)
    yield
  finally:
    StopTeeLogsToFile()



# Copyright 2014-2019 Chris Cummins <chrisc.101@gmail.com>.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Logging interface.
"""
from __future__ import print_function

import json

from labm8 import system


def colourise(colour, *args):
  return ''.join([colour] + list(args) + [Colours.RESET])


def printf(colour, *args, **kwargs):
  string = colourise(colour, *args)
  print(string, **kwargs)


def pprint(data, **kwargs):
  print(json.dumps(data, sort_keys=True, indent=2, separators=(',', ': ')),
        **kwargs)


def info(*args, **kwargs):
  print('[INFO  ]', *args, **kwargs)


def debug(*args, **kwargs):
  print('[DEBUG ]', *args, **kwargs)


def warn(*args, **kwargs):
  print('[WARN  ]', *args, **kwargs)


def error(*args, **kwargs):
  print('[ERROR ]', *args, **kwargs)


def fatal(*args, **kwargs):
  returncode = kwargs.pop('status', 1)
  error('fatal:', *args, **kwargs)
  system.exit(returncode)


def prof(*args, **kwargs):
  """
  Print a profiling message.

  Profiling messages are intended for printing runtime performance
  data. They are prefixed by the "PROF" title.

  Arguments:

      *args, **kwargs: Message payload.
  """
  print('[PROF  ]', *args, **kwargs)


class Colours:
  """
  Shell escape colour codes.
  """
  RESET = '\033[0m'
  GREEN = '\033[92m'
  YELLOW = '\033[93m'
  BLUE = '\033[94m'
  RED = '\033[91m'



"""A module for working within Bazel's hemetic and sandboxed world.

The design goals of Bazel has led it to have a lot of quirks. This module
contains utility code designed to help live with these quirks.
"""
import os
import pathlib
import re
import subprocess
import typing

from labm8 import app
from labm8 import archive
from labm8 import fs

# Regular expression to everything in a path up until the '*.runfiles'
# directory, e.g. for the path '/private/var/bazel/foo/bar.runfiles/a/b/c',
# this regex will match '/private/var/bazel/foo/bar.runfiles'
RUNFILES_PATTERN = re.compile(r'^(.*\.runfiles)/')


def FindRunfilesDirectory() -> typing.Optional[pathlib.Path]:
  """Find the '.runfiles' directory, if there is one.

  Returns:
    The absolute path of the runfiles directory, else None if not found.
  """
  # Follow symlinks, looking for my module space
  stub_filename = os.path.abspath(__file__)
  module_space = stub_filename + '.runfiles'
  if os.path.isdir(module_space):
    return pathlib.Path(module_space)
  match = RUNFILES_PATTERN.match(os.path.abspath(__file__))
  if match:
    return pathlib.Path(match.group(1))
  return None


def DataPath(
    path: typing.Union[str, pathlib.Path],
    must_exist: bool = True,
) -> pathlib.Path:
  """Return the absolute path to a data file.

  This allows you to access files from the 'data' attribute of a Python
  target in Bazel. This is needed because the path to the file changes depending
  on whether the current process is executing with a 'bazel run' environment, or
  as a 'bazel-bin' script.

  Args:
    path: The path to the data, including the name of the workspace.
    must_exist: Require that the file exists, else raise FileNotFoundError.

  Returns:
    An absolute file path.

  Raises:
    FileNotFoundError: If the requested path is not found and must_exist is
      True.
  """
  if not str(path):
    if must_exist:
      raise FileNotFoundError(f"No such file or directory: '{path}'")
    else:
      # An empty path yields the runfiles directory.
      return FindRunfilesDirectory()
  runfiles = FindRunfilesDirectory()
  real_path = runfiles / path if runfiles else pathlib.Path(path).absolute()
  if must_exist and not (real_path.is_file() or real_path.is_dir()):
    raise FileNotFoundError(f"No such file or directory: '{path}'")
  return real_path


def DataString(path: typing.Union[str, pathlib.Path]) -> str:
  """Return the contents of a data file.

  This allows you to access files from the 'data' attribute of a Python
  target in Bazel. This is needed because the path to the file changes depending
  on whether the current process is executing with a 'bazel run' environment, or
  as a 'bazel-bin' script.

  Args:
    path: The path to the data, including the name of the workspace.

  Returns:
    The contents of the file

  Raises:
    FileNotFoundError: If the requested path is not found.
  """
  with open(DataPath(path)) as f:
    contents = f.read()
  return contents


class DataArchive(archive.Archive):
  """A compressed archive file.

  TODO: data=[...] attribute of target.

  Provides access to an unzipped data file when used as a context manager by
  extracting the zip contents to a temporary directory.

  Example:
    >>> with DataArchive("phd/data.zip") as uncompressed_root:
    ...   print(uncompressed_root.iterdir())
    ['a', 'README.txt']
  """

  def __init__(self, path: typing.Union[str, pathlib.Path]):
    """Constructor.

    Args:
      path: The path to the data, including the name of the workspace.

    Raises:
      FileNotFoundError: If path is not a file.
    """
    super(DataArchive, self).__init__(DataPath(path))


class Workspace(object):
  """Class representing a bazel workspace."""

  def __init__(self, root: pathlib.Path):
    """Create a bazel workspace object.

    The workspace must already exist, this object does not modify files.

    Args:
       root: The root of the workspace.

    Raises:
      OSError: If the root is not a workspace.
    """
    self._root = root
    if not (self._root / 'WORKSPACE').is_file():
      raise OSError(f'`{self._root}/WORKSPACE` not found')

  @property
  def workspace_root(self) -> pathlib.Path:
    return self._root

  def BazelQuery(self,
                 args: typing.List[str],
                 timeout_seconds: int = 360,
                 **subprocess_kwargs):
    """Run bazel query with the specified args in the workspace.

    Args:
      args: The list of arguments to pass to bazel query.
      timeout_seconds: The number of seconds before failing.
      subprocess_kwargs: Additional arguments to pass to Popen().
    """
    return self.Bazel('query',
                      args,
                      timeout_seconds=timeout_seconds,
                      **subprocess_kwargs)

  def Bazel(self,
            command: str,
            args: typing.List[str],
            timeout_seconds: int = 360,
            **subprocess_kwargs):
    cmd = [
        'timeout',
        '-s9',
        str(timeout_seconds),
        'bazel',
        command,
        '--noshow_progress',
    ] + args
    app.Log(2, '$ %s', ' '.join(cmd))
    with fs.chdir(self.workspace_root):
      return subprocess.Popen(cmd, **subprocess_kwargs)

  def MaybeTargetToPath(
      self,
      fully_qualified_target: str,
  ) -> typing.Optional[pathlib.Path]:
    """Determine if a bazel target refers to a file, and if so return the path.

    Args:
      fully_qualified_target: The bazel target, beginning with '//'.

    Returns:
      A path, relative to the root of the workspace, else NOne.

    Raises:
      ValueError: If the given target is not fully qualified.
    """

    def RelpathIfExists(path: str) -> typing.Optional[pathlib.Path]:
      """Return if given relative path is a file."""
      abspath = self.workspace_root / path
      return path if abspath.is_file() else None

    if fully_qualified_target.startswith('//:'):
      return RelpathIfExists(fully_qualified_target[3:])
    elif fully_qualified_target.startswith('//'):
      return RelpathIfExists(fully_qualified_target[2:].replace(':', '/'))
    else:
      raise ValueError(
          'Target is not fully qualified (does not begin with `//`): '
          f'{fully_qualified_target}',)

  def GetDependentFiles(
      self,
      target: str,
      excluded_targets: typing.Iterable[str],
  ) -> typing.List[pathlib.Path]:
    """Get the file dependencies of the target.

    Args:
      target: The target to get the dependencies of.

    Returns:
      A list of paths, relative to the root of the workspace.

    Raises:
      OSError: If bazel query fails.
    """
    # First run through bazel query to expand globs.
    bazel = self.BazelQuery([target], stdout=subprocess.PIPE)
    grep = subprocess.Popen(
        ['grep', '^/'],
        stdout=subprocess.PIPE,
        stdin=bazel.stdout,
        universal_newlines=True,
    )

    stdout, _ = grep.communicate()
    if bazel.returncode:
      raise OSError('bazel query failed')
    if grep.returncode:
      raise OSError('grep of bazel query output failed')
    targets = stdout.rstrip().split('\n')

    # Now get the transitive dependencies of each target.
    targets = [target for target in targets if target not in excluded_targets]
    all_targets = targets.copy()
    for i, target in enumerate(targets):
      app.Log(1, 'Collecting transitive deps for target %d of %d: %s', i + 1,
              len(targets), target)
      bazel = self.BazelQuery([f'deps({target})'], stdout=subprocess.PIPE)
      grep = subprocess.Popen(
          ['grep', '^/'],
          stdout=subprocess.PIPE,
          stdin=bazel.stdout,
          universal_newlines=True,
      )

      stdout, _ = grep.communicate()
      if bazel.returncode:
        raise OSError('bazel query failed')
      if grep.returncode:
        raise OSError('grep of bazel query output failed')

      deps = stdout.rstrip().split('\n')
      all_targets += [
          target for target in deps if target not in excluded_targets
      ]

    paths = [self.MaybeTargetToPath(target) for target in all_targets]
    return [path for path in paths if path]

  def GetBuildFiles(self, target: str) -> typing.List[pathlib.Path]:
    """Get the BUILD files required for the given target.

    Raises:
      OSError: If bazel query fails.
    """
    bazel = self.BazelQuery(
        [f'buildfiles(deps({target}))'],
        stdout=subprocess.PIPE,
    )
    cut = subprocess.Popen(
        ['cut', '-f1', '-d:'],
        stdout=subprocess.PIPE,
        stdin=bazel.stdout,
    )
    grep = subprocess.Popen(
        ['grep', '^/'],
        stdout=subprocess.PIPE,
        stdin=cut.stdout,
        universal_newlines=True,
    )

    stdout, _ = grep.communicate()
    if bazel.returncode:
      raise OSError('bazel query failed')
    if cut.returncode:
      raise OSError('bazel query output cut failed')
    if grep.returncode:
      raise OSError('bazel query output search failed')

    for line in stdout.rstrip().split('\n'):
      if line == '//external':
        # Files in //external are virtual.
        continue
      path = os.path.join(line[2:], 'BUILD')
      abspath = self.workspace_root / path
      if not abspath.is_file():
        raise OSError(f'BUILD file not found: {path}')
      yield path



# Copyright 2014-2019 Chris Cummins <chrisc.101@gmail.com>.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A wrapper around absl's app and logging modules.

See: <https://github.com/abseil/abseil-py>
"""
import json
import sys

import functools
import pathlib
import re
from absl import app as absl_app
from absl import flags as absl_flags
from absl import logging as absl_logging
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from labm8 import shell
from labm8.internal import flags_parsers
from labm8.internal import logging


FLAGS = absl_flags.FLAGS

absl_flags.DEFINE_boolean(
    'version',
    False,
    'Print version information and exit.',
)
absl_flags.DEFINE_boolean(
    'dump_flags',
    False,
    'Print the defined flags and their values and exit.')
absl_flags.DEFINE_boolean(
    'dump_flags_to_json',
    False,
    'Print the defined flags and their values to JSON and exit.')
absl_flags.DEFINE_boolean('log_colors', True,
                          'Whether to colorize logging output.')


class UsageError(absl_app.UsageError):
  """Exception raised when the arguments supplied by the user are invalid.
  Raise this when the arguments supplied are invalid from the point of
  view of the application. For example when two mutually exclusive
  flags have been supplied or when there are not enough non-flag
  arguments.
  """

  def __init__(self, message, exitcode=1):
    super(UsageError, self).__init__(message)
    self.exitcode = exitcode


def AssertOrRaise(stmt: bool, exception: Exception, *exception_args,
                  **exception_kwargs) -> None:
  """If the statement is false, raise the given exception class."""
  if not stmt:
    raise exception(*exception_args, **exception_kwargs)


def GetVersionInformationString() -> str:
  """Return a string of version information, as printed by --version flag."""
  # If this is a bazel environment, then the //:build_info package will be
  # available. However, if this is a labm8 pip package install, then
  # //:build_info will not be available, so use pkg_resources to get the
  # version information.
  try:
    import build_info
    version = '\n'.join([
      build_info.FormatVersion(),
      build_info.FormatShortBuildDescription(),
    ])
    url = build_info.GetGithubCommitUrl()
  except ModuleNotFoundError:
    import pkg_resources
    version = f'version: {pkg_resources.get_distribution("labm8").version}'
    url = 'https://github.com/ChrisCummins/labm8'
  return '\n'.join([
    version,
    'Copyright (C) 2014-2019 Chris Cummins <chrisc.101@gmail.com>',
    f'<{url}>',
  ])


def RunWithArgs(
    main: Callable[[List[str]], None],
    argv: Optional[List[str]] = None,
):
  """Begin executing the program.

  Args:
    main: The main function to execute. It takes an single argument "argv",
      which is a list of command line arguments with parsed flags removed.
      If it returns an integer, it is used as the process's exit code.
    argv: A non-empty list of the command line arguments including program name,
      sys.argv is used if None.
  """

  def DoMain(argv):
    """Run the user-provided main method, with app-level arg handling."""
    if FLAGS.version:
      print(GetVersionInformationString())
      sys.exit(0)
    elif FLAGS.dump_flags:
      print(FlagsToString())
      sys.exit(0)
    elif FLAGS.dump_flags_to_json:
      flags_dict = FlagsToDict()
      # Flags values can have non-serializable types, so try each one and
      # stringify those that require it. An alternative would be to stringify
      # all values, but this would lose type information on ints/floats/etc.
      for flag in flags_dict:
        try:
          json.dumps(flags_dict[flag])
        except TypeError:
          flags_dict[flag] = str(flags_dict[flag])
      print(json.dumps(flags_dict, sort_keys=True, indent=2,
                       separators=(',', ': ')))
      sys.exit(0)
    main(argv)

  try:
    absl_app.run(DoMain, argv=argv)
  except KeyboardInterrupt:
    FlushLogs()
    sys.stdout.flush()
    sys.stderr.flush()
    print('keyboard interrupt')
    sys.exit(1)


def Run(main: Callable[[], None]):
  """Begin executing the program.

  Args:
    main: The main function to execute. It takes no arguments. If any command
    line arguments remain after flags parsing, an error is raised. If it
    returns an integer, it is used as the process's exit code.
  """

  def RunWithoutArgs(argv: List[str]):
    """Run the given function without arguments."""
    if len(argv) > 1:
      raise UsageError("Unknown arguments: '{}'.".format(' '.join(argv[1:])))
    main()

  RunWithArgs(RunWithoutArgs)


# Logging functions.


def GetVerbosity() -> int:
  """Get the verbosity level.

  This can be set per-module using --vmodule flag.
  """
  return logging.GetVerbosity(logging.GetCallingModuleName())


def _MaybeColorizeLog(color: str, msg: str, *args) -> str:
  """Conditionally apply shell colorization to the given format string."""
  string = str(msg) % args
  if FLAGS.log_colors:
    return f"{shell.ShellEscapeCodes.BOLD}{color}{string}{shell.ShellEscapeCodes.END}"
  else:
    return string


# Skip this function when determining the calling module and line number for
# logging.
@absl_logging.skip_log_prefix
def Log(level: int, msg, *args, **kwargs):
  """Logs a message at the given level.

  Per-module verbose level. The argument has to contain a comma-separated
  list of <module name>=<log level>. <module name> is a glob pattern (e.g., "
    "gfs* for all modules whose name starts with \"gfs\"), matched against the "
    "filename base (that is, name ignoring .py). <log level> overrides any "
    "value given by --v."
  """
  calling_module = logging.GetCallingModuleName()
  logging.Log(
      calling_module, level,
      _MaybeColorizeLog(
          shell.ShellEscapeCodes.YELLOW
          if level > 1 else shell.ShellEscapeCodes.CYAN, msg, *args), **kwargs)


@absl_logging.skip_log_prefix
def LogIf(level: int, condition, msg, *args, **kwargs):
  if condition:
    calling_module = logging.GetCallingModuleName()
    Log(level, msg, *args, **kwargs)


@absl_logging.skip_log_prefix
def Fatal(msg, *args, **kwargs):
  """Logs a fatal message."""
  logging.Fatal(
      _MaybeColorizeLog(shell.ShellEscapeCodes.RED, msg, *args), **kwargs)


@absl_logging.skip_log_prefix
def FatalWithoutStackTrace(msg, *args, **kwargs):
  """Logs a fatal message without stacktrace."""
  Error(msg, *args, **kwargs)
  sys.exit(1)


@absl_logging.skip_log_prefix
def Error(msg, *args, **kwargs):
  """Logs an error message."""
  logging.Error(
      _MaybeColorizeLog(shell.ShellEscapeCodes.RED, msg, *args), **kwargs)


@absl_logging.skip_log_prefix
def Warning(msg, *args, **kwargs):
  """Logs a warning message."""
  logging.Warning(
      _MaybeColorizeLog(shell.ShellEscapeCodes.RED, msg, *args), **kwargs)


def FlushLogs():
  """Flushes all log files."""
  logging.FlushLogs()


# TODO(cec): Consider emoving DebugLogging() in favour of GetVerbosity().
def DebugLogging() -> bool:
  """Return whether debug logging is enabled."""
  return logging.DebugLogging()


def SetLogLevel(level: int) -> None:
  """Sets the logging verbosity.

  Causes all messages of level <= v to be logged, and all messages of level > v
  to be silently discarded.

  Args:
    level: the verbosity level as an integer.
  """
  logging.SetLogLevel(level)


# Flags functions.

# This is a set of module ids for the modules that disclaim key flags.
# This module is explicitly added to this set so that we never consider it to
# define key flag.
disclaim_module_ids = set([id(sys.modules[__name__])])


@functools.lru_cache(maxsize=1)
def get_main_module_name(abspath) -> str:
  # Strip everything until the runfiles directory.
  name = re.sub(r'.*\.runfiles/[^/]+/', '', abspath)
  # Strip file extension.
  name = '.'.join(name.split('.')[:-1])
  # Change path separator to python module separator.
  name = '.'.join(name.split('/'))
  # Prefix name with a space so that it come first in the list of modules.
  return f' {name}'


def get_module_object_and_name(globals_dict):
  """Returns the module that defines a global environment, and its name.
  Args:
    globals_dict: A dictionary that should correspond to an environment
      providing the values of the globals.
  Returns:
    _ModuleObjectAndName - pair of module object & module name.
    Returns (None, None) if the module could not be identified.
  """
  name = globals_dict.get('__name__', None)
  module = sys.modules.get(name, None)
  # Pick a more informative name for the main module.
  return module, (sys.argv[0] if name == '__main__' else name)


def get_calling_module_name():
  """Returns the module that's calling into this module.
  We generally use this function to get the name of the module calling a
  DEFINE_foo... function.
  Returns:
    The module name that called into this one.
  Raises:
    AssertionError: Raised when no calling module could be identified.
  """
  for depth in range(1, sys.getrecursionlimit()):
    # sys._getframe is the right thing to use here, as it's the best
    # way to walk up the call stack.
    globals_for_frame = sys._getframe(depth).f_globals  # pylint: disable=protected-access
    module, module_name = get_module_object_and_name(globals_for_frame)
    if id(module) not in disclaim_module_ids and module_name is not None:
      return module_name
  raise AssertionError('No module was found')


# TODO(cec): Add flag_values argument to enable better testing.
# TODO(cec): Add validator callbacks.


def FlagsToDict() -> Dict[str, Any]:
  """Return a dictionary of flags and their values.

  Keys are the names of flags, prefixed by their defining module, e.g.
  "absl.flags.alsologtosterr" refers to flag "alsologtosterr" to module
  "absl.flags". Values are the string values in their defined types.

  Returns:
    A <flag, value> dictionary.
  """
  flags_dict = FLAGS.flags_by_module_dict()
  flattened_flags_dict = {}
  for module in flags_dict:
    for flag in flags_dict[module]:
      flattened_flags_dict[f'{module}.{flag.name}'] = flag.value
  return flattened_flags_dict

def FlagsToString() -> str:
  """Return the defined flags as a string.

  The string returned by this method is suitable to be used as a flagfile.

  Returns:
    A string of newline-separated flag values, in the form "--someflag=val".
  """
  return FLAGS.flags_into_string()


def DEFINE_string(
    name: str,
    default: Optional[str],
    help: str,
    required: bool = False,
    validator: Callable[[str], bool] = None,
):
  """Registers a flag whose value can be any string."""
  absl_flags.DEFINE_string(
      name,
      default,
      help,
      module_name=get_calling_module_name(),
  )
  if required:
    absl_flags.mark_flag_as_required(name)
  if validator:
    RegisterFlagValidator(name, validator)


def DEFINE_integer(
    name: str,
    default: Optional[int],
    help: str,
    required: bool = False,
    lower_bound: Optional[int] = None,
    upper_bound: Optional[int] = None,
    validator: Callable[[int], bool] = None,
):
  """Registers a flag whose value must be an integer."""
  absl_flags.DEFINE_integer(
      name,
      default,
      help,
      module_name=get_calling_module_name(),
      lower_bound=lower_bound,
      upper_bound=upper_bound,
  )
  if required:
    absl_flags.mark_flag_as_required(name)
  if validator:
    RegisterFlagValidator(name, validator)


def DEFINE_float(
    name: str,
    default: Optional[float],
    help: str,
    required: bool = False,
    lower_bound: Optional[float] = None,
    upper_bound: Optional[float] = None,
    validator: Callable[[float], bool] = None,
):
  """Registers a flag whose value must be a float."""
  absl_flags.DEFINE_float(
      name,
      default,
      help,
      module_name=get_calling_module_name(),
      lower_bound=lower_bound,
      upper_bound=upper_bound,
  )
  if required:
    absl_flags.mark_flag_as_required(name)
  if validator:
    RegisterFlagValidator(name, validator)


def DEFINE_boolean(
    name: str,
    default: Optional[bool],
    help: str,
    required: bool = False,
    validator: Callable[[bool], bool] = None,
):
  """Registers a flag whose value must be a boolean."""
  absl_flags.DEFINE_boolean(
      name,
      default,
      help,
      module_name=get_calling_module_name(),
  )
  if required:
    absl_flags.mark_flag_as_required(name)
  if validator:
    RegisterFlagValidator(name, validator)


def DEFINE_list(
    name: str,
    default: Optional[List[Any]],
    help: str,
    required: bool = False,
    validator: Callable[[List[Any]], bool] = None,
):
  """Registers a flag whose value must be a list."""
  absl_flags.DEFINE_list(
      name,
      default,
      help,
      module_name=get_calling_module_name(),
  )
  if required:
    absl_flags.mark_flag_as_required(name)
  if validator:
    RegisterFlagValidator(name, validator)


# My custom flag types.


def DEFINE_input_path(
    name: str,
    default: Union[None, str, pathlib.Path],
    help: str,
    required: bool = False,
    is_dir: bool = False,
    validator: Callable[[pathlib.Path], bool] = None,
):
  """Registers a flag whose value is an input path.

  An "input path" is a path to a file or directory that exists. The parsed value
  is a pathlib.Path instance. Flag parsing will fail if the value of this flag
  is not a path to an existing file or directory.

  Args:
    name: The name of the flag.
    default: The default value for the flag. While None is a legal value, it
      will fail during parsing - input paths are required flags.
    help: The help string.
    is_dir: If true, require the that the value be a directory. Else, require
      that the value be a file. Parsing will fail if this is not the case.
  """
  parser = flags_parsers.PathParser(must_exist=True, is_dir=is_dir)
  serializer = absl_flags.ArgumentSerializer()
  absl_flags.DEFINE(
      parser,
      name,
      default,
      help,
      absl_flags.FLAGS,
      serializer,
      module_name=get_calling_module_name(),
  )
  if required:
    absl_flags.mark_flag_as_required(name)
  if validator:
    RegisterFlagValidator(name, validator)


def DEFINE_output_path(
    name: str,
    default: Union[None, str, pathlib.Path],
    help: str,
    required: bool = False,
    is_dir: bool = False,
    exist_ok: bool = True,
    must_exist: bool = False,
    validator: Callable[[pathlib.Path], bool] = None,
):
  """Registers a flag whose value is an output path.

  An "output path" is a path to a file or directory that may or may not already
  exist. The parsed value is a pathlib.Path instance. The idea is that this flag
  can be used to specify paths to files or directories that will be created
  during program execution. However, note that specifying an output path does
  not guarantee that the file will be produced.

  Args:
    name: The name of the flag.
    default: The default value for the flag. While None is a legal value, it
      will fail during parsing - output paths are required flags.
    help: The help string.
    is_dir: If true, require the that the value be a directory. Else, require
      that the value be a file. Parsing will fail if the path already exists and
      is of the incorrect type.
    exist_ok: If False, require that the path not exist, else parsing will fail.
    must_exist: If True, require that the path exists, else parsing will fail.
  """
  parser = flags_parsers.PathParser(
      must_exist=must_exist,
      exist_ok=exist_ok,
      is_dir=is_dir,
  )
  serializer = absl_flags.ArgumentSerializer()
  absl_flags.DEFINE(
      parser,
      name,
      default,
      help,
      absl_flags.FLAGS,
      serializer,
      module_name=get_calling_module_name(),
  )
  if required:
    absl_flags.mark_flag_as_required(name)
  if validator:
    RegisterFlagValidator(name, validator)


def DEFINE_database(
    name: str,
    database_class,
    default: Optional[str],
    help: str,
    must_exist: bool = False,
    validator: Callable[[Any], bool] = None,
):
  """Registers a flag whose value is a sqlutil.Database class.

  Unlike other DEFINE_* functions, the value produced by this flag is not an
  instance of the value, but a lambda that will instantiate a database of the
  requested type. This flag value must be called (with no arguments) in order to
  instantiate a database.

  Args:
    name: The name of the flag.
    database_class: The subclass of sqlutil.Database which is to be instantiated
      when this value is called, using the URL declared in 'default'.
    default: The default URL of the database. This is a required value.
    help: The help string.
    must_exist: If True, require that the database exists. Else, the database is
      created if it does not exist.
  """
  parser = flags_parsers.DatabaseParser(database_class, must_exist=must_exist)
  serializer = absl_flags.ArgumentSerializer()
  absl_flags.DEFINE(
      parser,
      name,
      default,
      help,
      absl_flags.FLAGS,
      serializer,
      module_name=get_calling_module_name(),
  )
  if validator:
    RegisterFlagValidator(name, validator)


def RegisterFlagValidator(
    flag_name: str,
    checker: Callable[[Any], bool],
    message: str = 'Flag validation failed',
):
  """Adds a constraint, which will be enforced during program execution.

  The constraint is validated when flags are initially parsed, and after each
  change of the corresponding flag's value.

  Args:
    flag_name: str, name of the flag to be checked.
    checker: callable, a function to validate the flag.
        input - A single positional argument: The value of the corresponding
            flag (string, boolean, etc.  This value will be passed to checker
            by the library).
        output - bool, True if validator constraint is satisfied.
            If constraint is not satisfied, it should either return False or
            raise flags.ValidationError(desired_error_message).
    message: str, error text to be shown to the user if checker returns False.
        If checker raises flags.ValidationError, message from the raised
        error will be shown.

  Raises:
    AttributeError: Raised when flag_name is not registered as a valid flag
        name.
  """
  absl_flags.register_validator(flag_name, checker, message)


def LogToDirectory(logdir: Union[str, pathlib.Path],
                   name='info') -> pathlib.Path:
  """Write logs to a directory.

  This disables printing of logs to stderr, unless the --alsologtostderr flag
  is provided.

  Args:
    logdir: The directory to write logs to. This is created if it does not
      exist.
    name: The name of the log file.
  """
  logdir = pathlib.Path(logdir)
  logdir.mkdir(exist_ok=True, parents=True)
  absl_logging.get_absl_handler().use_absl_log_file(str(name), str(logdir))
  return logdir



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: labm8/internal/logging.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='labm8/internal/logging.proto',
  package='',
  syntax='proto2',
  serialized_options=None,
  serialized_pb=_b('\n\x1clabm8/internal/logging.proto\"\x9a\x02\n\tLogRecord\x12)\n\x05level\x18\x01 \x01(\x0e\x32\x13.LogRecord.LogLevelR\x05level\x12)\n\x11\x64\x61te_utc_epoch_ms\x18\x02 \x01(\x03R\x0e\x64\x61teUtcEpochMs\x12\x1b\n\tthread_id\x18\x03 \x01(\x03R\x08threadId\x12\x1b\n\tfile_name\x18\x04 \x01(\tR\x08\x66ileName\x12\x1f\n\x0bline_number\x18\x05 \x01(\x05R\nlineNumber\x12\x18\n\x07message\x18\x06 \x01(\tR\x07message\"B\n\x08LogLevel\x12\t\n\x05\x44\x45\x42UG\x10\x00\x12\x08\n\x04INFO\x10\x01\x12\x0b\n\x07WARNING\x10\x02\x12\t\n\x05\x45RROR\x10\x03\x12\t\n\x05\x46\x41TAL\x10\x04')
)



_LOGRECORD_LOGLEVEL = _descriptor.EnumDescriptor(
  name='LogLevel',
  full_name='LogRecord.LogLevel',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='DEBUG', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INFO', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='WARNING', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ERROR', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FATAL', index=4, number=4,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=249,
  serialized_end=315,
)
_sym_db.RegisterEnumDescriptor(_LOGRECORD_LOGLEVEL)


_LOGRECORD = _descriptor.Descriptor(
  name='LogRecord',
  full_name='LogRecord',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='level', full_name='LogRecord.level', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='level', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='date_utc_epoch_ms', full_name='LogRecord.date_utc_epoch_ms', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='dateUtcEpochMs', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='thread_id', full_name='LogRecord.thread_id', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='threadId', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='file_name', full_name='LogRecord.file_name', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='fileName', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='line_number', full_name='LogRecord.line_number', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='lineNumber', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='message', full_name='LogRecord.message', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='message', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _LOGRECORD_LOGLEVEL,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=33,
  serialized_end=315,
)

_LOGRECORD.fields_by_name['level'].enum_type = _LOGRECORD_LOGLEVEL
_LOGRECORD_LOGLEVEL.containing_type = _LOGRECORD
DESCRIPTOR.message_types_by_name['LogRecord'] = _LOGRECORD
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

LogRecord = _reflection.GeneratedProtocolMessageType('LogRecord', (_message.Message,), dict(
  DESCRIPTOR = _LOGRECORD,
  __module__ = 'labm8.internal.logging_pb2'
  # @@protoc_insertion_point(class_scope:LogRecord)
  ))
_sym_db.RegisterMessage(LogRecord)


# @@protoc_insertion_point(module_scope)



# Copyright 2014-2019 Chris Cummins <chrisc.101@gmail.com>.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility code for working with shells."""
import os



class ShellEscapeCodes(object):
  """Shell escape codes for pretty-printing."""
  PURPLE = '\033[95m'
  CYAN = '\033[96m'
  DARKCYAN = '\033[36m'
  BLUE = '\033[94m'
  GREEN = '\033[92m'
  YELLOW = '\033[93m'
  RED = '\033[91m'
  BOLD = '\033[1m'
  UNDERLINE = '\033[4m'
  END = '\033[0m'


def ShellEscapeList(words):
  """Turn a list of words into a shell-safe string.

  Args:
    words: A list of words, e.g. for a command.

  Returns:
    A string of shell-quoted and space-separated words.
  """
  # Adapted from <https://github.com/google/google-apputils>.
  # Copyright 2007 Google Inc. All Rights Reserved.
  #
  # Licensed under the Apache License, Version 2.0 (the "License");
  # you may not use this file except in compliance with the License.
  # You may obtain a copy of the License at
  #
  #      http://www.apache.org/licenses/LICENSE-2.0
  #
  # Unless required by applicable law or agreed to in writing, software
  # distributed under the License is distributed on an "AS-IS" BASIS,
  # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  # See the License for the specific language governing permissions and
  # limitations under the License.
  if os.name == 'nt':
    return ' '.join(words)

  s = ''
  for word in words:
    # Single quote word, and replace each ' in word with '"'"'
    s += "'" + word.replace("'", "'\"'\"'") + "' "

  return s[:-1]



# Copyright 2014-2019 Chris Cummins <chrisc.101@gmail.com>.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Useful function decorators."""
import functools

import contextlib
import signal
import typing

from labm8 import app

FLAGS = app.FLAGS

# A type hint that specifies a callable function with any number of arguments
# and any return type.
AnyFunction = typing.Callable[..., typing.Any]


def memoized_property(func: AnyFunction) -> AnyFunction:
  """A property decorator that memoizes the result.

  This is used to memoize the results of class properties, to be used when
  computing the property value is expensive.

  Args:
    func: The function which should be made to a property.

  Returns:
    The decorated property function.
  """
  # Based on Danijar Hafner's implementation of a lazy property, available at:
  # https://danijar.com/structuring-your-tensorflow-models/
  attribute_name = '_memoized_property_' + func.__name__

  @property
  @functools.wraps(func)
  def decorator(self):
    if not hasattr(self, attribute_name):
      setattr(self, attribute_name, func(self))
    return getattr(self, attribute_name)

  return decorator


@contextlib.contextmanager
def timeout(seconds: int):
  """A function decorator that raises TimeoutError after specified time limit.

  Args:
    seconds: The number of seconds before timing out.

  Raises:
    TimeoutError: If the number of seconds have been reached.
  """

  def _RaiseTimoutError(signum, frame):
    raise TimeoutError(f'Function failed to complete within {seconds} seconds')

  # Register a function to raise a TimeoutError on the signal.
  signal.signal(signal.SIGALRM, _RaiseTimoutError)
  signal.alarm(seconds)

  try:
    yield
  except TimeoutError as e:
    raise e
  finally:
    # Unregister the signal so it won't be triggered
    # if the timeout is not reached.
    signal.signal(signal.SIGALRM, signal.SIG_IGN)


@contextlib.contextmanager
def timeout_without_exception(seconds: int):
  """A function decorator that adds a timeout.

  Args:
    seconds: The number of seconds before timing out.
  """

  def _RaiseTimoutError(signum, frame):
    raise TimeoutError

  # Register a function to raise a TimeoutError on the signal.
  signal.signal(signal.SIGALRM, _RaiseTimoutError)
  signal.alarm(seconds)

  try:
    yield
  except TimeoutError:
    pass
  finally:
    # Unregister the signal so it won't be triggered
    # if the timeout is not reached.
    signal.signal(signal.SIGALRM, signal.SIG_IGN)


def run_once(f):
  """Runs a function (successfully) only once.

  The running can be reset by setting the `has_run` attribute to False

  Author: Jason Grout.
  From: https://gist.github.com/jasongrout/3804691
  """

  @functools.wraps(f)
  def wrapper(*args, **kwargs):
    if not wrapper.has_run:
      wrapper.result = f(*args, **kwargs)
      wrapper.has_run = True
    return wrapper.result

  wrapper.has_run = False
  return wrapper



# Copyright 2014-2019 Chris Cummins <chrisc.101@gmail.com>.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Wrapper for invoking Makefiles.
"""
import re

from labm8 import fs
from labm8 import system


class Error(Exception):
  """
  Module-level error class.
  """
  pass


class MakeError(Error):
  """
  Thrown if make command fails.
  """
  pass


class NoMakefileError(Error):
  """
  Thrown if a directory does not contain a Makefile.
  """
  pass


class NoTargetError(Error):
  """
  Thrown if there is no rule for the requested target.
  """
  pass


_BAD_TARGET_RE = re.compile(r'No rule to make target ' "'.+'.  Stop.")


def make(target='all', dir='.', **kwargs):
  """
  Run make.

  Arguments:

      target (str, optional): Name of the target to build. Defaults
        to "all".
      dir (str, optional): Path to directory containing Makefile.
      **kwargs (optional): Any additional arguments to be passed to
        system.run().

  Returns:

      (int, str, str): The first element is the return code of the
        make command. The second and third elements are the stdout
        and stderr of the process.

  Raises:

      NoMakefileError: In case a Makefile is not found in the target
        directory.
      NoTargetError: In case the Makefile does not support the
        requested target.
      MakeError: In case the target rule fails.
  """
  if not fs.isfile(fs.path(dir, 'Makefile')):
    raise NoMakefileError("No makefile in '{}'".format(fs.abspath(dir)))

  fs.cd(dir)

  # Default parameters to system.run()
  if 'timeout' not in kwargs: kwargs['timeout'] = 300

  ret, out, err = system.run(['make', target], **kwargs)
  fs.cdpop()

  if ret > 0:
    if re.search(_BAD_TARGET_RE, err):
      raise NoTargetError("No rule for target '{}'".format(target))
    else:
      raise MakeError("Target '{}' failed".format(target))

    raise MakeError('Failed')

  return ret, out, err


def clean(**kwargs):
  """
  Run make clean.

  Equivalent to invoking make() with target="clean".

  Arguments:

      **kwargs (optional): Any additional arguments to be passed to
        make().

  Returns:

      (int, str, str): The first element is the return code of the
        make command. The second and third elements are the stdout
        and stderr of the process.

  Raises:

      NoMakefileError: In case a Makefile is not found in the target
        directory.
      NoTargetError: In case the Makefile does not support the
        requested target.
      MakeError: In case the target rule fails.
  """
  make(target='clean', **kwargs)



# Copyright 2014-2019 Chris Cummins <chrisc.101@gmail.com>.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""String formatting utils.
"""
import typing

import pandas as pd


class Error(Exception):
  """
  Module-level error.
  """
  pass


def IndentList(num_spaces: int, strings: typing.List[str]) -> typing.List[str]:
  """Indent each string in a list of strings by a number of spaces.

  Args:
      num_spaces: The number of spacces to indent by. Must be greater than or
        equal to zero.
      strings: The list of strings to indent.

  Returns:
    A list of indented strings.
  """
  prefix = ''.join([' '] * num_spaces)
  return [f'{prefix}{string}' for string in strings]


def Indent(num_spaces: int, text: str) -> str:
  """Indent a string by a number of spaces.

  Prepends 'num_spaces' whitespace characters to every line in the given string.

  Args:
    num_spaces: The number of spaces to indent by. Must be greater than or equal
      to zero.
    text: The string to indent.

  Returns:
    The indented string.
  """
  return '\n'.join(IndentList(num_spaces, text.split('\n')))


def table(rows, columns=None, output=None, data_args={}, **kwargs) -> str:
  """
  Return a formatted string of "list of list" table data.

  See: http://pandas.pydata.org/pandas-docs/dev/generated/pandas.DataFrame.html

  Examples:
    >>> table([("foo", 1), ("bar", 2)])
         0  1
    0  foo  1
    1  bar  2

    >>> table([("foo", 1), ("bar", 2)], columns=("type", "value"))
      type  value
    0  foo      1
    1  bar      2

  Arguments:
    rows (list of list): Data to format, one row per element,
      multiple columns per row.
    columns (list of str, optional): Column names.
    output (str, optional): Path to output file.
    data_args (dict, optional): Any additional kwargs to pass to
      pd.DataFrame constructor.
    **kwargs: Any additional arguments to pass to
      pd.DataFrame.to_string().

  Returns:
    str: Formatted data as table.

  Raises:
    Error: If number of columns (if provided) does not equal
      number of columns in rows; or if number of columns is not
      consistent across all rows.
  """
  # Number of columns.
  num_columns = len(rows[0])

  # Check that each row is the same length.
  for i, row in enumerate(rows[1:]):
    if len(row) != num_columns:
      raise Error(
          'Number of columns in row {i_row} ({c_row}) '
          'does not match number of columns in row 0 ({z_row})'.format(
              i_row=i,
              c_row=len(row),
              z_row=num_columns,
          ),)

  if columns is None:
    # Default parameters.
    if 'header' not in kwargs:
      kwargs['header'] = False
  elif len(columns) != num_columns:
    # Check that number of columns matches number of columns in
    # rows.
    raise Error(
        'Number of columns in header ({c_header}) does not '
        'match the number of columns in the data ({c_rows})'.format(
            c_header=len(columns),
            c_rows=num_columns,
        ),)

  # Default arguments.
  if 'index' not in kwargs:
    kwargs['index'] = False

  data_args['columns'] = columns

  string = pd.DataFrame(list(rows), **data_args).to_string(**kwargs)
  if output is None:
    return string
  else:
    print(string, file=open(output, 'w'))
    print('Wrote', output)



# Copyright 2014-2019 Chris Cummins <chrisc.101@gmail.com>.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Lock file mechanism."""
import datetime
import inspect
import os
import pathlib
import sys
import time
import typing

from labm8 import app
from labm8 import humanize
from labm8 import labdate
from labm8 import pbutil
from labm8 import system
from labm8.internal import lockfile_pb2
# Use absolute paths for imports so as to prevent a conflict with the
# system "time" module.

FLAGS = app.FLAGS

app.DEFINE_float(
    'lockfile_block_seconds',
    10.0,
    'The number of seconds to block for when waiting for a lock '
    'files.',
)


class Error(Exception):
  pass


class UnableToAcquireLockError(Error):
  """ thrown if cannot acquire lock """

  def __init__(self, lock: 'LockFile'):
    self.lock = lock

  def __str__(self):
    return f"""\
Unable to acquire file lock owned by a different process.
Lock acquired by process {self.lock.pid} on \
{self.lock.user}@{self.lock.hostname} at {self.lock.date}.

If you believe that this is an error and that no other
process holds the lock, you may remove the lock file:

   {self.lock.path}"""


class MalformedLockfileError(UnableToAcquireLockError):
  """ thrown if lockfile is malformed """

  def __init__(self, path: pathlib.Path):
    self.path = path

  def __str__(self):
    return f'Lock file is malformed: {self.path}'


class UnableToReleaseLockError(Error):
  """ thrown if cannot release lock """

  def __init__(self, lock: 'LockFile'):
    self.lock = lock

  def __str__(self):
    return f"""\
Unable to release file lock owned by a different process.
Lock acquired by process {self.lock.pid} on \
{self.lock.user}@{self.lock.hostname} at {self.lock.date}.

If you believe that this is an error and that no other
process holds the lock, you may remove the lock file:

   {self.lock.path}"""


class LockFile:
  """A lock file.

  Attributes:
    path: Path of lock file.
  """

  def __init__(self, path: typing.Union[str, pathlib.Path]):
    """Create a new directory lock.

    Args:
      path: Path to lock file.
    """
    self.path = pathlib.Path(path).expanduser().absolute()

  @property
  def pid(self) -> typing.Optional[datetime.datetime]:
    """The process ID of the lock. Value is None if lock is not claimed."""
    lockfile = self.read(self.path)
    if lockfile.HasField('owner_process_id'):
      return lockfile.owner_process_id
    else:
      return None

  @property
  def date(self) -> typing.Optional[datetime.datetime]:
    """The date that the lock was acquired. Value is None if lock is unclaimed.
    """
    lockfile = self.read(self.path)
    if lockfile.date_acquired_utc_epoch_ms:
      return labdate.DatetimeFromMillisecondsTimestamp(
          lockfile.date_acquired_utc_epoch_ms,)
    else:
      return None

  @property
  def hostname(self) -> typing.Optional[str]:
    """The hostname of the lock owner. Value is None if lock is unclaimed."""
    lockfile = self.read(self.path)
    if lockfile.HasField('owner_hostname'):
      return lockfile.owner_hostname
    else:
      return None

  @property
  def user(self) -> typing.Optional[str]:
    """The username of the lock owner. Value is None if lock is unclaimed."""
    lockfile = self.read(self.path)
    if lockfile.HasField('owner_user'):
      return lockfile.owner_user
    else:
      return None

  @property
  def islocked(self) -> bool:
    """Whether the directory is locked."""
    return self.path.is_file()

  @property
  def owned_by_self(self):
    """
    Whether the directory is locked by the current process.
    """
    return self.hostname == system.HOSTNAME and self.pid == os.getpid()

  def acquire(
      self,
      replace_stale: bool = False,
      force: bool = False,
      pid: int = None,
      block: bool = False,
  ) -> 'LockFile':
    """Acquire the lock.

    A lock can be claimed if any of these conditions are true:
      1. The lock is unheld by anyone.
      2. The lock is held but the 'force' argument is set.
      3. The lock is held by the current process.

    Args:
      replace_stale: If true, lock can be aquired from stale processes. A stale
        process is one which currently owns the parent lock, but no process with
        that PID is alive. A lock which is owned by a different hostname is
        never stale, since we cannot determine if the PID of a remote system is
        alive.
      force: If true, ignore any existing lock. If false, fail if lock already
        claimed.
      pid: If provided, force the process ID of the lock to this value.
        Otherwise the ID of the current process is used.
      block: If True, block indefinitely until the lock is available. Use with
        care!

    Returns:
      Self.

    Raises:
      UnableToAcquireLockError: If the lock is already claimed
        (not raised if force option is used).
    """

    def _create_lock():
      lockfile = lockfile_pb2.LockFile(
          owner_process_id=os.getpid() if pid is None else pid,
          owner_process_argv=' '.join(sys.argv),
          date_acquired_utc_epoch_ms=labdate.MillisecondsTimestamp(
              labdate.GetUtcMillisecondsNow(),),
          owner_hostname=system.HOSTNAME,
          owner_user=system.USERNAME,
      )
      pbutil.ToFile(lockfile, self.path, assume_filename='LOCK.pbtxt')

    while True:
      if self.islocked:
        lock_owner_pid = self.pid
        if self.owned_by_self:
          pass  # don't replace existing lock
          break
        elif force:
          _create_lock()
          break
        elif (replace_stale and self.hostname == system.HOSTNAME and
              not system.isprocess(lock_owner_pid)):
          _create_lock()
          break
        elif not block:
          raise UnableToAcquireLockError(self)
        # Block and try again later.
        app.Log(
            1,
            'Blocking on lockfile %s for %s seconds',
            self.path,
            humanize.Duration(FLAGS.lockfile_block_seconds),
        )
        time.sleep(FLAGS.lockfile_block_seconds)
      else:  # new lock
        _create_lock()
        break
    return self

  def release(self, force=False):
    """Release lock.

    To release a lock, we must already own the lock.

    Args:
      force: If true, ignore any existing lock owner.

    Raises:
      UnableToReleaseLockError: If the lock is claimed by another process (not
        raised if force option is used).
    """
    # There's no lock, so do nothing.
    if not self.islocked:
      return

    if self.owned_by_self or force:
      os.remove(self.path)
    else:
      raise UnableToReleaseLockError(self)

  def __repr__(self):
    return str(self.path)

  def __enter__(self):
    return self.acquire(replace_stale=True, block=True)

  def __exit__(self, type, value, tb):
    self.release()

  @staticmethod
  def read(path: typing.Union[str, pathlib.Path]) -> lockfile_pb2.LockFile:
    """Read the contents of a LockFile.

    Args:
      path: Path to lockfile.

    Returns:
      A LockFile proto.
    """
    path = pathlib.Path(path)
    if path.is_file():
      try:
        return pbutil.FromFile(
            path,
            lockfile_pb2.LockFile(),
            assume_filename='LOCK.pbtxt',
        )
      except pbutil.DecodeError:
        raise MalformedLockfileError(path)
    else:
      return lockfile_pb2.LockFile()


class AutoLockFile(LockFile):
  """A lockfile which derives it's path automatically.

  This LockFile suclass automatically derives its path from its position in
  the calling code. The path is:

    <root>/<calling_module>_<calling_method>_<calling_lineno>.pbtxt

  Where `root` is a root directory (by default, /tmp/phd/labm8/autolockfiles),
  `calling_file` is the name of the module containing the calling code,
  `calling_method` is the name of the method containing the calling code, and
  `calling_lineno` is the line number of the calling code.

  Use this class to conveniently mark sections of code that can only be executed
  by a single process at a time, e.g.

      def MyFunc():
        mutex = AutoLockFile()
        with mutex:
          DoSomeWork()

  Now, if MyFunc is invoked by two separate processes, one of them will fail.
  """

  def __init__(
      self,
      root: typing.Union[str, pathlib.Path] = '/tmp/phd/labm8/autolockfiles',
      granularity: str = 'line',
  ):
    """Constructor

    Args:
      root: The directory to place the lock file in.
      granularity: The granularity of the lock, one of {line,function,module}.
        'line' means that the lock is unique to the calling line. 'function'
        means that the lock is unique to the calling function. 'module' means
        that the lock is unique to the calling module.
    """
    root = pathlib.Path(root)

    # Unlike the regular LockFile, an AutoLockFile will create the necessary
    # parent directory.
    root.mkdir(parents=True, exist_ok=True)

    # Inspect the calling code to get the lockfile path components.
    frame = inspect.stack()[1]
    module_name = pathlib.Path(frame.filename).stem
    function_name = frame.function
    lineno = frame.lineno

    if granularity == 'line':
      path = root / f'{module_name}_{function_name}_{lineno}.pbtxt'
    elif granularity == 'function':
      path = root / f'{module_name}_{function_name}.pbtxt'
    elif granularity == 'module':
      path = root / f'{module_name}.pbtxt'
    else:
      raise TypeError(
          f"Invalid granularity '{granularity}'. Must be one of: "
          f'{{line,function,module}}',)

    super(AutoLockFile, self).__init__(path)



# Copyright 2014-2019 Chris Cummins <chrisc.101@gmail.com>.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""High level filesystem interface.
"""
import contextlib
import os.path
import pathlib
import re
import shutil
import tempfile
import typing
from glob import iglob

from send2trash import send2trash

from labm8 import humanize
from labm8 import labtypes


class Error(Exception):
  pass


class File404(Error):
  pass


def path(*components):
  """
  Get a file path.

  Concatenate all components into a path.
  """
  _path = os.path.join(*components)
  _path = os.path.expanduser(_path)

  return _path


def must_exist(*components):
  """
  Ensure path exists.

  Arguments:
      *components (str[]): Path components.

  Returns:
      str: File path.

  Raises:
      File404: If path does not exist.
  """
  _path = path(*components)
  if not exists(_path):
    raise File404(_path)
  return _path


def abspath(*components):
  """
  Get an absolute file path.

  Concatenate all components into an absolute path.
  """
  return os.path.abspath(path(*components))


def basename(*components):
  """
  Return the basename of a given file path.
  """
  return os.path.basename(path(*components))


def dirname(*components):
  """
  Return the directory name of a given file path.
  """
  return os.path.dirname(path(*components))


def is_subdir(child, parent):
  """
  Determine if "child" is a subdirectory of "parent".

  If child == parent, returns True.
  """
  child_path = os.path.realpath(child)
  parent_path = os.path.realpath(parent)

  if len(child_path) < len(parent_path):
    return False

  for i in range(len(parent_path)):
    if parent_path[i] != child_path[i]:
      return False

  return True


# Directory history.
_cdhist = []


def cd(path):
  """
  Change working directory.

  Returns absolute path to new working directory.
  """
  _cdhist.append(pwd())  # Push to history.
  path = abspath(path)
  os.chdir(path)
  return path


def cdpop():
  """
  Return the last directory.

  Returns absolute path to new working directory.
  """
  if len(_cdhist) >= 1:
    old = _cdhist.pop()  # Pop from history.
    os.chdir(old)
    return old
  else:
    return pwd()


def pwd():
  """
  Return the path to the current working directory.
  """
  return os.getcwd()


def exists(*components):
  """
  Return whether a file exists.
  """
  return os.path.exists(path(*components))


def isfile(*components):
  """
  Return whether a path exists, and is a file.
  """
  return os.path.isfile(path(*components))


def isexe(*components):
  """
  Return whether a path is an executable file.

  Arguments:
    path (str): Path of the file to check.

  Examples:
    >>> isexe("/bin/ls")
    True

    >>> isexe("/home")
    False

    >>> isexe("/not/a/real/path")
    False

  Returns:
      bool: True if file is executable, else false.
  """
  _path = path(*components)
  return isfile(_path) and os.access(_path, os.X_OK)


def isdir(*components):
  """
  Return whether a path exists, and is a directory.
  """
  if components:
    return os.path.isdir(path(*components))
  else:
    return False


def ls(
    root: typing.Union[str, pathlib.Path] = '.',
    abspaths=False,
    recursive=False,
):
  """
  Return a list of files in directory.

  Directory listings are sorted alphabetically. If the named
  directory is a file, return it's path.

  Examples:
    >>> ls("foo")
    ["a", "b", "c"]

    >>> ls("foo/a")
    ["foo/a"]

    >>> ls("foo", abspaths=True)
    ["/home/test/foo/a", "/home/test/foo/b", "/home/test/foo/c"]

    >>> ls("foo", recursive=True)
    ["a", "b", "b/d", "b/d/e", "c"]

  Arguments:
    root (str): Path to directory. Can be relative or absolute.
    abspaths (bool, optional): Return absolute paths if true.
    recursive (bool, optional): Recursively list subdirectories if
      true.

  Returns:
    list of str: A list of paths.

  Raises:
    OSError: If root directory does not exist.
  """

  def _expand_subdirs(file):
    if isdir(path(root, file)):
      return [
          file,
      ] + [path(file, x) for x in ls(path(root, file), recursive=True)]
    else:
      return [file]

  if isfile(root):
    # If argument is a file, return path.
    return [abspath(root)] if abspaths else [basename(root)]
  elif abspaths:
    # Get relative names.
    relpaths = ls(root, recursive=recursive, abspaths=False)
    # Prepend the absolute path to each relative name.
    base = abspath(root)
    return [path(base, relpath) for relpath in relpaths]
  elif recursive:
    # Recursively expand subdirectories.
    paths = ls(root, abspaths=abspaths, recursive=False)
    return labtypes.flatten([_expand_subdirs(file) for file in paths])
  else:
    # List directory contents.
    return list(sorted(os.listdir(root)))


def lsdirs(root='.', **kwargs):
  """
  Return only subdirectories from a directory listing.

  Arguments:

      root (str): Path to directory. Can be relative or absolute.
      **kwargs: Any additional arguments to be passed to ls().

  Returns:

      list of str: A list of directory paths.

  Raises:

      OSError: If root directory does not exist.
  """
  paths = ls(root=root, **kwargs)
  if isfile(root):
    return []
  return [_path for _path in paths if isdir(path(root, _path))]


def lsfiles(root: typing.Union[str, pathlib.Path] = '.', **kwargs):
  """
  Return only files from a directory listing.

  Arguments:

      root (str): Path to directory. Can be relative or absolute.
      **kwargs: Any additional arguments to be passed to ls().

  Returns:

      list of str: A list of file paths.

  Raises:

      OSError: If root directory does not exist.
  """
  paths = ls(root=root, **kwargs)
  if isfile(root):
    return paths
  return [_path for _path in paths if isfile(path(root, _path))]


def rm(*components, **kwargs):
  """
  Remove a file or directory.

  If path is a directory, this recursively removes the directory and
  any contents. Non-existent paths are silently ignored.

  Supports Unix style globbing by default (disable using
  glob=False). For details on globbing pattern expansion, see:

      https://docs.python.org/2/library/glob.html

  Arguments:
      *components (string[]): path to the file or directory to remove. May be
        absolute or relative. May contain unix glob
      **kwargs: if "glob" is True, perform Unix style pattern expansion of
        paths (default: True).
  """
  _path = path(*components)
  glob = kwargs.get('glob', True)
  paths = iglob(_path) if glob else [_path]

  for file in paths:
    if isfile(file):
      os.remove(file)
    elif exists(file):
      shutil.rmtree(file, ignore_errors=False)


def rmtrash(*components):
  """
  Move a file or directory to trash.

  If file does not exist, nothing happens.

  Examples:

      >>> rmtrash("foo", "bar")

      >>> rmtrash("/home/labm8/file.txt")

  Arguments:
      *components (string[]): path to the file or directory.
  """
  _path = path(*components)
  if exists(_path):
    send2trash(_path)


def cp(src, dst):
  """
  Copy a file or directory.

  If source is a directory, this recursively copies the directory
  and its contents. If the destination is a directory, then this
  creates a copy of the source in the destination directory with the
  same basename.

  If the destination already exists, this will attempt to overwrite
  it.

  Arguments:

      src (string): path to the source file or directory.
      dst (string): path to the destination file or directory.

  Raises:

      IOError: if source does not exist.
  """
  if isdir(src):
    # Overwrite an existing directory.
    if isdir(dst):
      rm(dst)
    shutil.copytree(src, dst)
  elif isfile(src):
    shutil.copy(src, dst)
  else:
    raise IOError("Source '{0}' not found".format(src))


def mv(src, dst):
  """
  Move a file or directory.

  If the destination already exists, this will attempt to overwrite
  it.

  Arguments:

      src (string): path to the source file or directory.
      dst (string): path to the destination file or directory.

  Raises:

      File404: if source does not exist.
      IOError: in case of error.
  """
  if not exists(src):
    raise File404(src)

  try:
    shutil.move(src, dst)
  except Exception as e:
    raise IOError(str(e))


def mkdir(*components, **kwargs):
  """
  Make directory "path", including any required parents. If
  directory already exists, do nothing.
  """
  _path = path(*components)
  if not isdir(_path):
    os.makedirs(_path, **kwargs)
  return _path


def mkopen(p, *args, **kwargs):
  """
  A wrapper for the open() builtin which makes parent directories if needed.
  """
  dir = os.path.dirname(p)
  mkdir(dir)
  return open(p, *args, **kwargs)


def read(*components, **kwargs):
  """
  Read file and return a list of lines. If comment_char is set, ignore the
  contents of lines following the comment_char.

  Raises:

      IOError: if reading path fails
  """
  rstrip = kwargs.get('rstrip', True)
  comment_char = kwargs.get('comment_char', None)

  ignore_comments = comment_char is not None

  file = open(path(*components))
  lines = file.readlines()
  file.close()

  # Multiple definitions to handle all cases.
  if ignore_comments:
    comment_line_re = re.compile('^\s*{char}'.format(char=comment_char))
    not_comment_re = re.compile('[^{char}]+'.format(char=comment_char))

    if rstrip:
      # Ignore comments, and right strip results.
      return [
          re.match(not_comment_re, line).group(0).rstrip()
          for line in lines
          if not re.match(comment_line_re, line)
      ]
    else:
      # Ignore comments, and don't strip results.
      return [
          re.match(not_comment_re, line).group(0)
          for line in lines
          if not re.match(comment_line_re, line)
      ]
  elif rstrip:
    # No comments, and right strip results.
    return [line.rstrip() for line in lines]
  else:
    # Just a good old-fashioned read!
    return lines


def du(*components, **kwargs):
  """
  Get the size of a file in bytes or as a human-readable string.

  Arguments:

      *components (str[]): Path to file.
      **kwargs: If "human_readable" is True, return a formatted string,
        e.g. "976.6 KiB" (default True)

  Returns:
      int or str: If "human_readble" kwarg is True, return str, else int.
  """
  human_readable = kwargs.get('human_readable', True)

  _path = path(*components)
  if not exists(_path):
    raise Error("file '{}' not found".format(_path))
  size = os.stat(_path).st_size
  if human_readable:
    return humanize.BinaryPrefix(size, 'B')
  else:
    return size


def files_from_list(*paths):
  """
  Return a list of all file paths from a list of files or directories.

  For each path in the input: if it is a file, return it; if it is a
  directory, return a list of files in the directory.

  Arguments:
      paths (list of str): List of file and directory paths.

  Returns:
      list of str: Absolute file paths.

  Raises:
      File404: If any of the paths do not exist.
  """
  ret = []
  for path in paths:
    if isfile(path):
      ret.append(abspath(path))
    elif isdir(path):
      ret += [f for f in ls(path, abspaths=True, recursive=True) if isfile(f)]
    else:
      raise File404(path)
  return ret


def directory_is_empty(directory: pathlib.Path) -> bool:
  """Return if a directory is empty.

  A directory which does not exist is considered empty (returns True). A
  directory containing only subdirectories but no files is considered not empty
  (returns False).

  Args:
    directory: The path of a directory.

  Returns:
    True if directory is empty, else False.
  """
  for _, subdirs, files in os.walk(path(directory)):
    if subdirs or files:
      return False
  return True


@contextlib.contextmanager
def chdir(directory: typing.Union[str, pathlib.Path]) -> pathlib.Path:
  """A context manager which allows you to temporarily change working directory.

  Args:
    directory: The directory to change to.

  Returns:
    The directory which has been changed to.

  Raises:
    OSError: If the given directory does not exist.
    NotADirectoryError: If the given path is a file.
  """
  previous_directory = pathlib.Path.cwd()
  os.chdir(str(directory))
  try:
    yield pathlib.Path(directory)
  finally:
    os.chdir(str(previous_directory))


@contextlib.contextmanager
def TemporaryWorkingDir(prefix: str = 'phd_') -> pathlib.Path:
  """A context manager which provides a temporary working directory.

  This creates an empty temporary directory, and changes the current working
  directory to it. Once out of scope, the directory and all it's contents are
  removed.

  Args:
    prefix: A prefix for the temporary directory name.

  Returns:
    The directory which has been changed to.
  """
  # getcwd() will raise FileNotFoundError if the current workind directory
  # does not exist.
  old_directory = None
  try:
    old_directory = os.getcwd()
  except FileNotFoundError:
    pass
  # Create a temporary directory, change to it, and return the path to the user.
  with tempfile.TemporaryDirectory(prefix=prefix) as d:
    os.chdir(d)
    yield pathlib.Path(d)
  # Return to previous working directory, if there was one.
  if old_directory:
    os.chdir(old_directory)


def Read(filename: typing.Union[str, pathlib.Path]) -> str:
  """Read entire contents of file with name 'filename'."""
  with open(filename) as fp:
    return fp.read()


def Write(
    filename: typing.Union[str, pathlib.Path],
    contents: bytes,
    overwrite_existing: bool = True,
    mode: int = 0o0666,
    gid: int = None,
) -> pathlib.Path:
  """Create a file 'filename' with 'contents', with the mode given in 'mode'.

  The 'mode' is modified by the umask, as in open(2).  If
  'overwrite_existing' is False, the file will be opened in O_EXCL mode.
  An optional gid can be specified.

  Args:
    filename: the name of the file
    contents: the data to write to the file
    overwrite_existing: whether or not to allow the write if the file
      already exists
    mode: permissions with which to create the file (default is 0666 octal)
    gid: group id with which to create the file
  """
  # Adapted from <https://github.com/google/google-apputils>.
  # Copyright 2007 Google Inc. All Rights Reserved.
  #
  # Licensed under the Apache License, Version 2.0 (the "License");
  # you may not use this file except in compliance with the License.
  # You may obtain a copy of the License at
  #
  #      http://www.apache.org/licenses/LICENSE-2.0
  #
  # Unless required by applicable law or agreed to in writing, software
  # distributed under the License is distributed on an "AS-IS" BASIS,
  # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  # See the License for the specific language governing permissions and
  # limitations under the License.
  flags = os.O_WRONLY | os.O_TRUNC | os.O_CREAT
  if not overwrite_existing:
    flags |= os.O_EXCL
  fd = os.open(filename, flags, mode)
  try:
    os.write(fd, contents)
  finally:
    os.close(fd)
  if gid is not None:
    os.chown(filename, -1, gid)
  return pathlib.Path(filename)


def AtomicWrite(
    filename: typing.Union[str, pathlib.Path],
    contents: bytes,
    mode: int = 0o0666,
    gid: int = None,
) -> None:
  """Create a file 'filename' with 'contents' atomically.

  As in Write, 'mode' is modified by the umask.  This creates and moves
  a temporary file, and errors doing the above will be propagated normally,
  though it will try to clean up the temporary file in that case.
  This is very similar to the prodlib function with the same name.
  An optional gid can be specified.

  Args:
    filename: the name of the file
    contents: the data to write to the file
    mode: permissions with which to create the file (default is 0666 octal)
    gid: group id with which to create the file
  """
  # Adapted from <https://github.com/google/google-apputils>.
  # Copyright 2007 Google Inc. All Rights Reserved.
  #
  # Licensed under the Apache License, Version 2.0 (the "License");
  # you may not use this file except in compliance with the License.
  # You may obtain a copy of the License at
  #
  #      http://www.apache.org/licenses/LICENSE-2.0
  #
  # Unless required by applicable law or agreed to in writing, software
  # distributed under the License is distributed on an "AS-IS" BASIS,
  # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  # See the License for the specific language governing permissions and
  # limitations under the License.
  fd, tmp_filename = tempfile.mkstemp(dir=os.path.dirname(filename))
  try:
    os.write(fd, contents)
  finally:
    os.close(fd)
  try:
    os.chmod(tmp_filename, mode)
    if gid is not None:
      os.chown(tmp_filename, -1, gid)
    os.rename(tmp_filename, filename)
  except OSError as exc:
    try:
      os.remove(tmp_filename)
    except OSError as e:
      exc = OSError('%s. Additional errors cleaning up: %s' % (exc, e))
    raise exc


@contextlib.contextmanager
def TemporaryFileWithContents(contents: bytes, **kwargs):
  """A contextmanager that writes out a string to a file on disk.

  This is useful whenever you need to call a function or command that expects a
  file on disk with some contents that you have in memory. The context manager
  abstracts the writing, flushing, and deletion of the temporary file. This is a
  common idiom that boils down to a single with statement.
  Note:  if you need a temporary file-like object for calling an internal
  function, you should use a StringIO as a file-like object and not this.
  Temporary files should be avoided unless you need a file name or contents in a
  file on disk to be read by some other function or program.

  Args:
    contents: a string with the contents to write to the file.
    **kwargs: Optional arguments passed on to tempfile.NamedTemporaryFile.

  Yields:
    The temporary file object, opened in 'w' mode.
  """
  # Adapted from <https://github.com/google/google-apputils>.
  # Copyright 2007 Google Inc. All Rights Reserved.
  #
  # Licensed under the Apache License, Version 2.0 (the "License");
  # you may not use this file except in compliance with the License.
  # You may obtain a copy of the License at
  #
  #      http://www.apache.org/licenses/LICENSE-2.0
  #
  # Unless required by applicable law or agreed to in writing, software
  # distributed under the License is distributed on an "AS-IS" BASIS,
  # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  # See the License for the specific language governing permissions and
  # limitations under the License.
  if not kwargs.get('prefix'):
    kwargs['prefix'] = 'phd_tempfile_with_contents_'
  temporary_file = tempfile.NamedTemporaryFile(**kwargs)
  temporary_file.write(contents)
  temporary_file.flush()
  yield temporary_file
  temporary_file.close()



# Copyright 2014-2019 Chris Cummins <chrisc.101@gmail.com>.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module defines classes and helper functions for working with graphs."""
import typing

from labm8 import app
from labm8 import fmt

FLAGS = app.FLAGS


class Graph(object):

  def __init__(
      self,
      name: str,
      children: typing.Optional[typing.List['Graph']] = None,
  ):
    self.name = name
    if children:
      self.children = set(children)
    else:
      self.children = set()

  def _ToDot(self, strings, visited: typing.Set['Graph']) -> None:
    if self in visited:
      return
    visited.add(self)
    strings.append(f'{self.name}[label="{self.name}"]')
    for child in self.children:
      strings.append(f'{self.name} -> {child.name}')
    for child in self.children:
      child._ToDot(strings, visited)

  def ToDot(self) -> str:
    strings = []
    self._ToDot(strings, set())
    dot = '\n'.join(fmt.IndentList(2, strings))
    return f'digraph graphname {{\n{dot}\n}}'

  def _PreOrderApply(
      self,
      callback: typing.Callable[['Graph'], None],
      visited: typing.Set['Graph'],
  ) -> None:
    if self in visited:
      return
    visited.add(self)
    for child in self.children:
      child._PreOrderApply(callback, visited)

  def PreOrderApply(self, callback: typing.Callable[['Graph'], None]) -> None:
    """Apply the given callback to all nodes in the graph.

    Args:
      callback: A callback function whose sole arguments accepts a Graph
        instance.
    """
    self._PreOrderApply(callback, set())

  def __eq__(self, other):
    return self.name == other.name

  def __neq__(self, other):
    return not self == other

  def __lt__(self, other):
    return self.name < other.name

  def __le__(self, other):
    return self < other or self == other

  def __hash__(self):
    return hash(self.name)

  def __repr__(self):
    return self.name



# Copyright 2014-2019 Chris Cummins <chrisc.101@gmail.com>.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Transient and persistent caching mechanisms."""
import atexit
import json
import pathlib
import re
import typing

import six

from labm8 import crypto
from labm8 import fs
from labm8 import io


# TODO(cec): Remove type hints on base Cache, place them on FSCache.
class Cache(object):
  """
  Cache for storing (key,value) relational data.

  A cache is a dictionary with a limited subset of a the
  functionality.
  """

  def get(self, key, default=None) -> typing.Optional[pathlib.Path]:
    """
    Retrieve an item from cache.

    Arguments:
        key: Item key.
        default (optional): Default value if item not found.
    """
    raise NotImplementedError

  def clear(self) -> None:
    """
    Remove all items from cache.
    """
    raise NotImplementedError

  def items(self) -> typing.Iterable[pathlib.Path]:
    """
    Returns a generator for iterating over (key, value) pairs.
    """
    raise NotImplementedError

  def __getitem__(self, key) -> pathlib.Path:
    """
    Retrieve an item from cache.

    Arguments:
       key: Item key.

    Raises:
       KeyError: If key is not in cache.
    """
    raise NotImplementedError

  def __setitem__(self, key, value) -> None:
    """
    Set (key, value) pair.
    """
    raise NotImplementedError

  def __contains__(self, key) -> bool:
    """
    Returns whether key is in cache.
    """
    raise NotImplementedError

  def __delitem__(self, key) -> None:
    """
    Remove (key, value) pair.
    """
    raise NotImplementedError

  def __iter__(self) -> typing.Iterator[pathlib.Path]:
    """
    Iterate over all cache entries.
    """
    raise NotImplementedError

  def __len__(self) -> int:
    """
    Get the number of entries in the cache.
    """
    raise NotImplementedError


class TransientCache(Cache):
  """
  An in-memory only cache.
  """

  def __init__(self, basecache=None):
    """
    Create a new transient cache.

    Optionally supports populating the cache with values of an
    existing cache.

    Arguments:
       basecache (TransientCache, optional): Cache to populate this new
         cache with.
    """
    self._data = {}

    if basecache is not None:
      for key, val in basecache.items():
        self._data[key] = val

  def get(self, key, default=None):
    if key in self._data:
      return self._data[key]
    else:
      return default

  def clear(self):
    self._data.clear()

  def items(self):
    return six.iteritems(self._data)

  def __getitem__(self, key):
    return self._data[key]

  def __setitem__(self, key, value):
    self._data[key] = value
    return value

  def __contains__(self, key):
    return key in self._data

  def __delitem__(self, key):
    del self._data[key]

  def __iter__(self):
    """
    Iterate over all cache entries.

    Returns:
        iterable: Entries in cache.
    """
    for value in self._data.values():
      yield value

  def __len__(self):
    """
    Get the number of cache entries.

    Returns:
        int: Number of entries in the cache.
    """
    return len(list(self._data.keys()))


class JsonCache(TransientCache):
  """
  A persistent, JSON-backed cache.

  Requires that (key, value) pairs are JSON serialisable.
  """

  def __init__(self, path, basecache=None):
    """
    Create a new JSON cache.

    Optionally supports populating the cache with values of an
    existing cache.

    Arguments:
       basecache (TransientCache, optional): Cache to populate this new
         cache with.
    """

    super(JsonCache, self).__init__()
    self.path = fs.abspath(path)

    if fs.exists(self.path) and fs.Read(self.path):
      io.debug(("Loading cache '{0}'".format(self.path)))
      with open(self.path) as file:
        self._data = json.load(file)

    if basecache is not None:
      for key, val in basecache.items():
        self._data[key] = val

    # Register exit handler
    atexit.register(self.write)

  def write(self):
    """
    Write contents of cache to disk.
    """
    io.debug("Storing cache '{0}'".format(self.path))
    with open(self.path, 'w') as file:
      json.dump(
          self._data,
          file,
          sort_keys=True,
          indent=2,
          separators=(',', ': '),
      )


def hash_key(key):
  """
  Convert a key to a filename by hashing its value.
  """
  return crypto.sha1_str(json.dumps(key, sort_keys=True))


def escape_path(key):
  """
  Convert a key to a filename by escaping invalid characters.
  """
  return re.sub(r'[ \\/]+', '_', key)


class FSCache(Cache):
  """
  Persistent filesystem cache.

  Each key uniquely identifies a file.
  Each value is a file path.

  Adding a file to the cache moves it into the cahce directory.

  Members:
      path (str): Root cache.
      escape_key (fn): Function to convert keys to file names.
  """

  def __init__(self, root, escape_key=hash_key):
    """
    Create filesystem cache.

    Arguments:
        root (str): String.
        escape_key (fn, optional): Function to convert keys to file names.
    """
    self.path = pathlib.Path(root)
    self.escape_key = escape_key

    fs.mkdir(self.path)

  def clear(self):
    """
    Empty the filesystem cache.

    This deletes the entire cache directory.
    """
    fs.rm(self.path)

  def keypath(self, key):
    """
    Get the filesystem path for a key.

    Arguments:
        key: Key.

    Returns:
        str: Absolute path.
    """
    return fs.path(self.path, self.escape_key(key))

  def __getitem__(self, key):
    """
    Get path to file in cache.

    Arguments:
        key: Key.

    Returns:
        str: Path to cache value.

    Raises:
        KeyErorr: If key not in cache.
    """
    path = self.keypath(key)
    if fs.exists(path):
      return path
    else:
      raise KeyError(key)

  def __setitem__(self, key, value):
    """
    Emplace file in cache.

    Arguments:
        key: Key.
        value (str): Path of file to insert in cache.

    Raises:
        ValueError: If no "value" does nto exist.
    """
    if not fs.exists(value):
      raise ValueError(value)

    path = self.keypath(key)
    fs.mkdir(self.path)
    fs.mv(value, path)

  def __contains__(self, key):
    """
    Check cache contents.

    Arguments:
        key: Key.

    Returns:
        bool: True if key in cache, else false.
    """
    path = self.keypath(key)
    return fs.exists(path)

  def __delitem__(self, key):
    """
    Delete cached file.

    Arguments:
        key: Key.

    Raises:
        KeyError: If file not in cache.
    """
    path = self.keypath(key)
    if fs.exists(path):
      fs.rm(path)
    else:
      raise KeyError(key)

  def __iter__(self):
    """
    Iterate over all cached files.

    Returns:
        iterable: Paths in cache.
    """
    for path in fs.ls(self.path, abspaths=True):
      yield path

  def __len__(self):
    """
    Get the number of entries in the cache.

    Returns:
        int: Number of entries in the cache.
    """
    return len(list(fs.ls(self.path)))

  def get(self, key, default=None):
    """
    Fetch from cache.

    Arguments:
        key: Key.
        default (optional): Value returned if key not found.

    Returns:
        str: Path to cached file.
    """
    if key in self:
      return self[key]
    else:
      return default

  def ls(self, **kwargs):
    """
    List files in cache.

    Arguments:
        **kwargs: Keyword options to pass to fs.ls().

    Returns:
        iterable: List of files.
    """
    return fs.ls(self.path, **kwargs)



# Copyright 2014-2019 Chris Cummins <chrisc.101@gmail.com>.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A hash cache for the filesystem.

Checksums files and directories and cache results. If a file or directory has
not been modified, subsequent hashes are cache hits. Hashes are recomputed
lazily, when a directory (or any of its subdirectories) have been modified.
"""
import collections
import os
import pathlib
import subprocess
import time
import typing

import checksumdir
import sqlalchemy as sql
from sqlalchemy.ext import declarative

from labm8 import app
from labm8 import crypto
from labm8 import fs
from labm8 import humanize
from labm8 import sqlutil

FLAGS = app.FLAGS

Base = declarative.declarative_base()

# An in-memory cache which is optionally shared amongst all HashCache instances.
# The in-memory cache omits timestamps from records.
InMemoryCacheKey = collections.namedtuple(
    'InMemoryCacheKey',
    ['hash_fn', 'path'],
)
IN_MEMORY_CACHE: typing.Dict[InMemoryCacheKey, str] = {}


class HashCacheRecord(Base):
  """A hashed file or directory."""
  __tablename__ = 'entries'

  # The absolute path to a file or directory.
  absolute_path: str = sql.Column(sql.String(4096), primary_key=True)
  # The number of seconds since the epoch that the file or directory was last
  # modified.
  last_modified: int = sql.Column(sql.Integer, nullable=False)
  # The cached hash in hexadecimal encoding. We use the length of the longest
  # supported hash function: sha256.
  hash: str = sql.Column(sql.String(64), nullable=False)


def GetDirectoryMTime(path: pathlib.Path) -> int:
  """Get the timestamp of the most recently modified file/dir in directory.

  Recursively checks subdirectory contents. This requires that the directory
  exists and is not empty.

  Params:
    abspath: The absolute path to the directory.

  Returns:
    The seconds since epoch of the last modification.
  """
  # Pure python implementation.
  # return int(max(
  #     max(os.path.getmtime(os.path.join(root, file)) for file in files) for
  #     root, _, files in os.walk(path)))
  # Faster implementation using UNIX tools. Requires GNU tools. On macOS, this
  # means having some homebrew package installed:
  #
  #    $ brew install findutils coreutils
  #
  # and the following directory in your PATH:
  #    /usr/local/opt/findutils/libexec/gnubin
  #    /usr/local/opt/coreutils/libexec/gnubin
  output = subprocess.check_output(
      f"find '{path}' -type f | xargs -d'\n' stat -c '%Y:%n' | sort -t: -n | "
      'tail -1 | cut -d: -f1',
      universal_newlines=True,
      shell=True,
  )
  return int(output)


class HashCache(sqlutil.Database):

  def __init__(
      self,
      path: pathlib.Path,
      hash_fn: str,
      keep_in_memory: bool = False,
  ):
    """Instantiate a hash cache.

    Args:
      path:
      hash_fn: The name of the hash function. One of: md5, sha1, sha256.
      keep_in_memory: If True, hashes are kept in memory for the lifespan
        of the process, or until Clear() is called on any HashCache instance.
        Use this with caution, as the in-memory cache does not invalidate
        entries, so cache entries can become stale.

    Raises:
      ValueError: If hash_fn not recognized.
    """
    super(HashCache, self).__init__(f'sqlite:///{path.absolute()}', Base)
    self.hash_fn_name = hash_fn
    if hash_fn == 'md5':
      self.hash_fn_file = crypto.md5_file
    elif hash_fn == 'sha1':
      self.hash_fn_file = crypto.sha1_file
    elif hash_fn == 'sha256':
      self.hash_fn_file = crypto.sha256_file
    else:
      raise ValueError(f"Hash function not recognized: '{hash_fn}'")
    self.keep_in_memory = keep_in_memory

  def GetHash(self, path: pathlib.Path) -> str:
    """Get the hash of a file or directory.

    This method is O(n) with respect to the number of files in the directory.
    The first time this is called for a directory, this method must read every
    file in the directory. For subsequent calls, this method must check the
    mtime of every file.

    Note that the a file's mtime is used to determine cache hits. This uses
    second granularity, so if a file has been modified within a second, this
    method will erroneously return the cached checksum of the previous version.

    Args:
      path: Path to the file or directory.

    Returns:
      Hexadecimal string hash.

    Raises:
      FileNotFoundError: If the requested path does not exist.
    """
    if path.is_file():
      return self._HashFile(path)
    elif path.is_dir():
      return self._HashDirectory(path)
    else:
      raise FileNotFoundError(f"File not found: '{path}'")

  def Clear(self):
    """Empty the cache.

    If the HashCache was created with keep_in_memory=True, this clears the
    in-memory cache. Note that the in-memory cache is shared between all
    instances of HashCache.
    """
    IN_MEMORY_CACHE.clear()
    with self.Session(commit=True) as session:
      session.query(HashCacheRecord).delete()
    app.Log(2, 'Emptied cache')

  def _HashDirectory(self, absolute_path: pathlib.Path) -> str:
    if fs.directory_is_empty(absolute_path):
      last_modified_fn = lambda path: int(time.time())
    else:
      last_modified_fn = lambda path: GetDirectoryMTime(path)
    return self._InMemoryWrapper(
        absolute_path,
        last_modified_fn,
        lambda x: checksumdir.dirhash(x, self.hash_fn_name),
    )

  def _HashFile(self, absolute_path: pathlib.Path) -> str:
    return self._InMemoryWrapper(
        absolute_path,
        lambda path: int(os.path.getmtime(path)),
        self.hash_fn_file,
    )

  def _InMemoryWrapper(
      self,
      absolute_path: pathlib.Path,
      last_modified_fn: typing.Callable[[pathlib.Path], int],
      hash_fn: typing.Callable[[pathlib.Path], str],
  ) -> str:
    """A wrapper around the persistent hashing to support in-memory cache."""
    if self.keep_in_memory:
      in_memory_key = InMemoryCacheKey(self.hash_fn_name, absolute_path)
      if in_memory_key in IN_MEMORY_CACHE:
        app.Log(2, "In-memory cache hit: '%s'", absolute_path)
        return IN_MEMORY_CACHE[in_memory_key]
    hash_ = self._DoHash(
        absolute_path,
        last_modified_fn(absolute_path),
        hash_fn,
    )
    if self.keep_in_memory:
      IN_MEMORY_CACHE[in_memory_key] = hash_
    return hash_

  def _DoHash(
      self,
      absolute_path: pathlib.Path,
      last_modified: int,
      hash_fn: typing.Callable[[pathlib.Path], str],
  ) -> str:
    with self.Session() as session:
      cached_entry = session.query(HashCacheRecord).filter(
          HashCacheRecord.absolute_path == str(absolute_path),).first()
      if cached_entry and cached_entry.last_modified == last_modified:
        app.Log(2, "Cache hit: '%s'", absolute_path)
        return cached_entry.hash
      elif cached_entry:
        app.Log(2, "Cache miss: '%s'", absolute_path)
        session.delete(cached_entry)
      start_time = time.time()
      checksum = hash_fn(absolute_path)
      app.Log(
          2,
          "New cache entry '%s' in %s ms.",
          absolute_path,
          humanize.Commas(int((time.time() - start_time) * 1000)),
      )
      new_entry = HashCacheRecord(
          absolute_path=str(absolute_path),
          last_modified=last_modified,
          hash=checksum,
      )
      session.add(new_entry)
      session.commit()
      return new_entry.hash



# Copyright 2014-2019 Chris Cummins <chrisc.101@gmail.com>.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""JSON parser which supports comments.
"""
import json
import re
import typing

from labm8 import fs

# A type alias for annotating methods which take or return JSON.
JSON = typing.Union[typing.List[typing.Any], typing.Dict[str, typing.Any]]


def format_json(data):
  """
  Pretty print JSON.

  Arguments:
      data (dict): JSON blob.

  Returns:
      str: Formatted JSON
  """
  return json.dumps(data, sort_keys=True, indent=2, separators=(',', ': '))


def read_file(*components, **kwargs):
  """
  Load a JSON data blob.

  Arguments:
      path (str): Path to file.
      must_exist (bool, otional): If False, return empty dict if file does
          not exist.

  Returns:
      array or dict: JSON data.

  Raises:
      File404: If path does not exist, and must_exist is True.
      InvalidFile: If JSON is malformed.
  """
  must_exist = kwargs.get('must_exist', True)

  if must_exist:
    path = fs.must_exist(*components)
  else:
    path = fs.path(*components)

  try:
    with open(path) as infile:
      return loads(infile.read())
  except ValueError as e:
    raise ValueError(
        "malformed JSON file '{path}'. Message from parser: {err}".format(
            path=fs.basename(path),
            err=str(e),
        ),)
  except IOError as e:
    if not must_exist:
      return {}
    else:
      return e


def write_file(path, data, format=True):
  """
  Write JSON data to file.

  Arguments:
      path (str): Destination.
      data (dict or list): JSON serializable data.
      format (bool, optional): Pretty-print JSON data.
  """
  if format:
    fs.Write(path, format_json(data).encode('utf-8'))
  else:
    fs.Write(path, json.dumps(data).encode('utf-8'))


def loads(text, **kwargs):
  """
  Deserialize `text` (a `str` or `unicode` instance containing a JSON
  document with Python or JavaScript like comments) to a Python object.

  Supported comment types: `// comment` and `# comment`.

  Taken from `commentjson <https://github.com/vaidik/commentjson>`_, written
  by `Vaidik Kapoor <https://github.com/vaidik>`_.

  Copyright (c) 2014 Vaidik Kapoor, MIT license.

  Arguments:
      text (str): serialized JSON string with or without comments.
      **kwargs (optional): all the arguments that
          `json.loads <http://docs.python.org/2/library/json.html#json.loads>`_
          accepts.

  Returns:
      `dict` or `list`: Decoded JSON.
  """
  regex = r'\s*(#|\/{2}).*$'
  regex_inline = r'(:?(?:\s)*([A-Za-z\d\.{}]*)|((?<=\").*\"),?)(?:\s)*(((#|(\/{2})).*)|)$'
  lines = text.split('\n')

  for index, line in enumerate(lines):
    if re.search(regex, line):
      if re.search(r'^' + regex, line, re.IGNORECASE):
        lines[index] = ''
      elif re.search(regex_inline, line):
        lines[index] = re.sub(regex_inline, r'\1', line)

  return json.loads('\n'.join(lines), **kwargs)



# Copyright 2014-2019 Chris Cummins <chrisc.101@gmail.com>.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Math utils. Import as "labmath" to prevent conflicts with system math package.
"""
# Use absolute paths for imports so as to prevent a conflict with the
# system "math" module.
from __future__ import absolute_import
from __future__ import division

import math

import numpy as np
from scipy import stats
# Use floating point division by default.


def ceil(number):
  """
  Return the ceiling of a number as an int.

  This is the smallest integral value >= number.

  Example:

      >>> ceil(1.5)
      2

  Arguments:

      number (float): A numeric value.

  Returns:

      int: Smallest integer >= number.

  Raises:

      TypeError: If argument is not a numeric value.
  """
  return int(math.ceil(number))


def floor(number):
  """
  Return the floor of a number as an int.

  This is the largest integral value <= number.

  Example:

      >>> floor(1.5)
      1

  Arguments:

      number (float): A numeric value.

  Returns:

      int: Largest integer <= number.

  Raises:

      TypeError: If argument is not a numeric value.
  """
  return int(math.floor(number))


def sqrt(number):
  """
  Return the square root of a number.
  """
  return math.sqrt(number)


def mean(array):
  """
  Return the mean value of a list of divisible numbers.
  """
  n = len(array)

  if n < 1:
    return 0
  elif n == 1:
    return array[0]
  return sum(array) / n


def geomean(array):
  """
  Return the mean value of a list of divisible numbers.
  """
  n = len(array)

  if n < 1:
    return 0
  elif n == 1:
    return array[0]
  return stats.mstats.gmean(array)


def median(array):
  """
  Return the median value of a list of numbers.
  """
  n = len(array)

  if n < 1:
    return 0
  elif n == 1:
    return array[0]

  sorted_vals = sorted(array)
  midpoint = int(n / 2)
  if n % 2 == 1:
    return sorted_vals[midpoint]
  else:
    return (sorted_vals[midpoint - 1] + sorted_vals[midpoint]) / 2.0


def range(array):
  """
  Return the range between min and max values.
  """
  if len(array) < 1:
    return 0
  return max(array) - min(array)


def variance(array):
  """
  Return the variance of a list of divisible numbers.
  """
  if len(array) < 2:
    return 0
  u = mean(array)
  return sum([(x - u)**2 for x in array]) / (len(array) - 1)


def stdev(array):
  """
  Return the standard deviation of a list of divisible numbers.
  """
  return sqrt(variance(array))


def iqr(array, lower, upper):
  """
  Return interquartile range for given array.

  Arguments:

      lower (float): Lower bound for IQR, in range 0 <= lower <= 1.
      upper (float): Upper bound for IQR, in range 0 <= upper <= 1.

  Returns:

      (float, float): Lower and upper IQR values.
  """
  return np.percentile(array, [lower * 100, upper * 100])


def filter_iqr(array, lower, upper):
  """
  Return elements which falls within specified interquartile range.

  Arguments:

      array (list): Sequence of numbers.
      lower (float): Lower bound for IQR, in range 0 <= lower <= 1.
      upper (float): Upper bound for IQR, in range 0 <= upper <= 1.

  Returns:

      list: Copy of original list, with elements outside of IQR
        removed.
  """
  upper, lower = iqr(array, upper, lower)

  new = list(array)
  for x in new[:]:
    if x < lower or x > upper:
      new.remove(x)

  return new


def confinterval(
    array,
    conf=0.95,
    normal_threshold=30,
    error_only=False,
    array_mean=None,
):
  """
  Return the confidence interval of a list for a given confidence.

  Arguments:

      array (list): Sequence of numbers.
      conf (float): Confidence interval, in range 0 <= ci <= 1
      normal_threshold (int): The number of elements in the array is
        < normal_threshold, use a T-distribution. If the number of
        elements in the array is >= normal_threshold, use a normal
        distribution.
      error_only (bool, optional): If true, return only the size of
        symmetrical confidence interval, equal to ci_upper - mean.
      array_mean (float, optional): Optimisation trick for if you
        already know the arithmetic mean of the array to prevent
        this function from re-calculating.

  Returns:

      (float, float): Lower and upper bounds on confidence interval,
        respectively.
  """
  n = len(array)

  if n < 1:
    # We have no data.
    array_mean, c0, c1 = 0, 0, 0
  elif n == 1:
    # We have only a single datapoint, so return that value.
    array_mean, c0, c1 = array[0], array[0], array[0]
  else:
    scale = stdev(array) / sqrt(n)
    # Check if all values are the same.
    values_all_same = all(x == array[0] for x in array[1:])

    if values_all_same:
      # If values are all the same, return that value.
      array_mean, c0, c1 = array[0], array[0], array[0]
    else:
      if array_mean is None: array_mean = mean(array)
      if n < normal_threshold:
        # We have a "small" number of datapoints, so use a
        # t-distribution.
        c0, c1 = stats.t.interval(conf, n - 1, loc=array_mean, scale=scale)
      else:
        # We have a "large" number of datapoints, so use a
        # normal distribution.
        c0, c1 = stats.norm.interval(conf, loc=array_mean, scale=scale)

  if error_only:
    return c1 - array_mean
  else:
    return c0, c1



# Copyright 2014-2019 Chris Cummins <chrisc.101@gmail.com>.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for generating LaTeX.
"""
from __future__ import print_function

import re
from io import StringIO

import pandas

from labm8 import io


class Error(Exception):
  """
  Module-level error.
  """
  pass


def escape(text):
  return re.sub(r'(_)', r'\\\g<1>', str(text))


def wrap_bold(text):
  return '\\textbf{' + text + '}'


def write_table_body(
    data,
    output=None,
    headers=None,
    header_fmt=wrap_bold,
    hline_after_header=True,
    hline_before=False,
    hline_after=False,
):

  def _write_row(row):
    output.write(' & '.join(escape(column) for column in row) + '\\\\\n')

  # Determine if we're writing to a file or returning a string.
  isfile = output is not None
  output = output or StringIO()

  # Write hline before body.
  if hline_before:
    output.write('\\hline\n')

  # Write headers.
  if headers:
    _write_row(header_fmt(str(column)) for column in headers)
    # Write hline after headers.
    if hline_after_header:
      output.write('\\hline\n')

  # Write table entries.
  for row in data:
    _write_row(row)

  # Write hline after body.
  if hline_after:
    output.write('\\hline\n')

  return None if isfile else output.getvalue()


def table(rows, columns=None, output=None, data_args={}, **kwargs):
  """
  Return a LaTeX formatted string of "list of list" table data.

  See: http://pandas.pydata.org/pandas-docs/dev/generated/pandas.DataFrame.html

  Requires the "booktabs" package to be included in LaTeX preamble:

      \\usepackage{booktabs}

  Examples:

      >>> fmt.print([("foo", 1), ("bar", 2)])
           0  1
      0  foo  1
      1  bar  2

      >>> fmt.print([("foo", 1), ("bar", 2)], columns=("type", "value"))
        type  value
      0  foo      1
      1  bar      2

  Arguments:

      rows (list of list): Data to format, one row per element,
        multiple columns per row.
      columns (list of str, optional): Column names.
      output (str, optional): Path to output file.
      data_args (dict, optional): Any additional kwargs to pass to
        pandas.DataFrame constructor.
      **kwargs: Any additional arguments to pass to
        pandas.DataFrame.to_latex().

  Returns:

      str: Formatted data as LaTeX table.

  Raises:

      Error: If number of columns (if provided) does not equal
        number of columns in rows; or if number of columns is not
        consistent across all rows.
  """
  # Number of columns.
  num_columns = len(rows[0])

  # Check that each row is the same length.
  for i, row in enumerate(rows[1:]):
    if len(row) != num_columns:
      raise Error(
          'Number of columns in row {i_row} ({c_row}) '
          'does not match number of columns in row 0 ({z_row})'.format(
              i_row=i,
              c_row=len(row),
              z_row=num_columns,
          ),)

  # Check that (if supplied), number of columns matches number of
  # columns in rows.
  if columns is not None and len(columns) != num_columns:
    raise Error(
        'Number of columns in header ({c_header}) does not '
        'match the number of columns in the data ({c_rows})'.format(
            c_header=len(columns),
            c_rows=num_columns,
        ),)

  # Default arguments.
  if 'index' not in kwargs:
    kwargs['index'] = False

  data_args['columns'] = columns

  string = pandas.DataFrame(list(rows), **data_args).to_latex(**kwargs)
  if output is None:
    return string
  else:
    print(string, file=open(output, 'w'))
    io.info('Wrote', output)



# Copyright 2014-2019 Chris Cummins <chrisc.101@gmail.com>.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This file defines date handling logic.

Frequently we wish to use UTC dates, with a maximum of millisecond precision.
Using the methods defined in this file to create and convert dates ensures
equivalency when converting between timestamp and datetime formats.
"""
import datetime
import typing

import pytz

UTC = pytz.UTC
US_PACIFIC = pytz.timezone('US/Pacific')


def GetUtcMillisecondsNow() -> datetime.datetime:
  """Return the current date to millisecond precision.

  This method strips the microseconds returned value, allowing for equivalency
  checks before and after conversion to millisecond timestamps.

  Returns:
    A datetime instance.
  """
  d = datetime.datetime.utcnow()
  # Strip the microseconds. Don't round them to the nearest millisecond, as
  # this may cause rounding up beyond the range of valid microsecond values.
  return d.replace(microsecond=int(d.microsecond / 1000) * 1000)


def MillisecondsTimestamp(date: typing.Optional[datetime.datetime] = None,
                         ) -> int:
  """Get the millisecond timestamp of a date.

  Args:
    date: A datetime instance. If not provided, GetUtcMillisecondsNow is used.

  Returns:
    The milliseconds since the epoch of this date.

  Raises:
    TypeError: If the argument is of incorrect type.
  """
  date = date or GetUtcMillisecondsNow()
  if not isinstance(date, datetime.datetime):
    raise TypeError('Date must be a datetime instance')
  return int(date.strftime('%s%f')[:-3])


def DatetimeFromMillisecondsTimestamp(timestamp: int = None,
                                     ) -> datetime.datetime:
  """Get the date of a millisecond timestamp.

  Args:
    timestamp: Milliseconds since the epoch. If not provided, or if value is
      zero, the current time is used.

  Returns:
    A datetime instance.

  Raises:
    TypeError: If the argument is of incorrect type.
    ValueError: If the argument is not a positive integer.
  """
  if not (isinstance(timestamp, int) or timestamp is None):
    raise TypeError('Timestamp must be an integer')
  if not timestamp:
    timestamp = MillisecondsTimestamp()
  if timestamp < 0:
    raise ValueError('Negative timestamp not allowed')
  return datetime.datetime.fromtimestamp(timestamp / 1000)



STRING = """\
# File: //config.proto
# Proto: BuildInfo
seconds_since_epoch: 1571330358
host: "florence"
user: "cec"
repo_dirty: true
git_commit: "8e18722724d42babd8df70712efb05362d52a9ed"
git_branch: "wip/graph"
git_tracking: "priv/wip/graph"
git_remote_url: "git@github.com:ChrisCummins/phd-priv.git"
git_commit_author: "Chris Cummins <chrisc.101@gmail.com>"
git_commit_seconds_since_epoch: 1571330166
"""



# Copyright 2014-2019 Chris Cummins <chrisc.101@gmail.com>.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tarball util.
"""
import tarfile

from labm8 import fs


def unpack_archive(*components, **kwargs) -> str:
  """
  Unpack a compressed archive.

  Arguments:
      *components (str[]): Absolute path.
      **kwargs (dict, optional): Set "compression" to compression type.
          Default: bz2. Set "dir" to destination directory. Defaults to the
          directory of the archive.

  Returns:
      str: Path to directory.
  """
  path = fs.abspath(*components)
  compression = kwargs.get('compression', 'bz2')
  dir = kwargs.get('dir', fs.dirname(path))

  fs.cd(dir)
  tar = tarfile.open(path, 'r:' + compression)
  tar.extractall()
  tar.close()
  fs.cdpop()

  return dir



# Copyright 2014-2019 Chris Cummins <chrisc.101@gmail.com>.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility code for working with pandas."""
import pandas as pd
import typing
from absl import flags as absl_flags

from labm8 import sqlutil

FLAGS = absl_flags.FLAGS


def QueryToDataFrame(session: sqlutil.Session,
                     query: sqlutil.Query) -> pd.DataFrame:
  """Read query results to a Pandas DataFrame.

  Args:
    session: A database session.
    query: The query to run.

  Returns:
    A Pandas DataFrame.
  """
  return pd.read_sql(query.statement, session.bind)


def ModelToDataFrame(
    session: sqlutil.Session,
    model,
    columns: typing.Optional[typing.List[str]] = None,
    query_identity=lambda q: q,
):
  """Construct and execute a query reads an object's fields to a dataframe.

  Args:
    session: A database session.
    model: A database mapped object.
    columns: A list of column names, where each element is a column mapped to
      the model. If not provided, all column names are used.
    query_identity: A function which takes the produced query and returns a
      query. Use this to implement filtering of the query results.

  Returns:
    A Pandas DataFrame with one column for each field.
  """
  columns = columns or ColumnNames(model)
  query = session.query(*[getattr(model, column) for column in columns])
  df = QueryToDataFrame(session, query_identity(query))
  df.columns = columns
  return df
